from collections import namedtuple

from eth_utils import add_0x_prefix

from common_utils_py.agreements.service_agreement_template import ServiceAgreementTemplate
from common_utils_py.agreements.service_types import ServiceTypes, ServiceTypesIndices
from common_utils_py.ddo.service import Service
from common_utils_py.utils.utilities import generate_prefixed_id
from common_utils_py.utils.utilities import to_checksum_addresses

Agreement = namedtuple('Agreement', ('template', 'conditions'))


class ServiceAgreement(Service):
    """Class representing a Service Agreement."""
    SERVICE_INDEX = 'index'
    AGREEMENT_TEMPLATE = 'serviceAgreementTemplate'
    SERVICE_ATTRIBUTES = 'attributes'
    SERVICE_CONDITIONS = 'conditions'
    SERVICE_ENDPOINT = 'serviceEndpoint'

    def __init__(self, attributes, service_agreement_template, service_endpoint=None,
                 service_type=None):
        """

        :param attributes: attributes
        :param service_agreement_template: ServiceAgreementTemplate instance
        :param service_endpoint: str URL to use for requesting service defined in this agreement
        :param service_type: str like ServiceTypes.ASSET_ACCESS
        """
        self.service_agreement_template = service_agreement_template
        values = dict()
        values[ServiceAgreementTemplate.TEMPLATE_ID_KEY] = self.template_id
        values['attributes'] = dict()
        values['attributes'] = attributes
        values['attributes']['serviceAgreementTemplate'] = service_agreement_template.__dict__['template']
        if service_type == ServiceTypes.ASSET_ACCESS:
            values['index'] = ServiceTypesIndices.DEFAULT_ACCESS_INDEX
            Service.__init__(self, service_endpoint,
                             ServiceTypes.ASSET_ACCESS,
                             values, ServiceTypesIndices.DEFAULT_ACCESS_INDEX)

        elif service_type == ServiceTypes.CLOUD_COMPUTE:
            values['index'] = ServiceTypesIndices.DEFAULT_COMPUTING_INDEX
            Service.__init__(self, service_endpoint,
                             ServiceTypes.CLOUD_COMPUTE,
                             values, ServiceTypesIndices.DEFAULT_COMPUTING_INDEX)

        elif service_type == ServiceTypes.DID_SALES:
            values['index'] = ServiceTypesIndices.DEFAULT_DID_SALES_INDEX
            Service.__init__(self, service_endpoint,
                             ServiceTypes.DID_SALES,
                             values, ServiceTypesIndices.DEFAULT_DID_SALES_INDEX)

        elif service_type == ServiceTypes.NFT_SALES:
            values['index'] = ServiceTypesIndices.DEFAULT_NFT_SALES_INDEX
            Service.__init__(self, service_endpoint,
                             ServiceTypes.NFT_SALES,
                             values, ServiceTypesIndices.DEFAULT_NFT_SALES_INDEX)

        elif service_type == ServiceTypes.NFT_ACCESS:
            values['index'] = ServiceTypesIndices.DEFAULT_NFT_ACCESS_INDEX
            Service.__init__(self, service_endpoint,
                             ServiceTypes.NFT_ACCESS,
                             values, ServiceTypesIndices.DEFAULT_NFT_ACCESS_INDEX)

        elif service_type == ServiceTypes.NFT721_ACCESS:
            values['index'] = ServiceTypesIndices.DEFAULT_NFT721_ACCESS_INDEX
            Service.__init__(self, service_endpoint,
                             ServiceTypes.NFT_ACCESS,
                             values, ServiceTypesIndices.DEFAULT_NFT721_ACCESS_INDEX)

        elif service_type == ServiceTypes.ASSET_ACCESS_PROOF:
            values['index'] = ServiceTypesIndices.DEFAULT_ACCESS_PROOF_INDEX
            Service.__init__(self, service_endpoint,
                             ServiceTypes.ASSET_ACCESS_PROOF,
                             values, ServiceTypesIndices.DEFAULT_ACCESS_PROOF_INDEX)

        else:
            raise ValueError(f'The service_type {service_type} is not currently supported.')

    def get_number_nfts(self):
        """
        Return the number of nfts

        :return: int
        """
        return int(self.get_param_value_by_name('_numberNfts'))

    def get_key_hash(self):
        """
        Return the key hash

        :return: Str
        """
        return self.attributes['main']['_hash']

    def get_provider_babyjub_key(self):
        """
        Return the provider babyjub key

        :return: Str[]
        """
        return self.attributes['main']['_providerPub']

    def get_nft_holder(self):
        """
        Return the NFT holder

        :return Str
        """
        return self.get_param_value_by_name('_nftHolder')

    def get_amounts(self):
        """
        Return the list of amounts/rewards to distribute

        :return: Str[]
        """
        return self.get_param_value_by_name('_amounts')

    def get_amounts_int(self):
        """
        Return the list of amounts/rewards to distribute

        :return: Int[]
        """
        return list(map(int, self.get_amounts()))

    def get_receivers(self):
        """
        Return the list of receivers addresses

        :return: str[]
        """
        return to_checksum_addresses(self.get_param_value_by_name('_receivers'))

    def get_price(self):
        """
        Return the price from the conditions parameters.

        :return: Int
        """
        price = 0
        _amounts = self.get_param_value_by_name('_amounts')
        for amount in _amounts:
            price = price + int(amount)
        return price

    def get_param_value_by_name(self, name):
        """
        Return the value from the conditions parameters given the param name.

        :return: Object
        """
        for cond in self.conditions:
            for p in cond.parameters:
                if p.name == name:
                    return p.value

    @property
    def service_endpoint(self):
        """

        :return:
        """
        return self._service_endpoint

    @property
    def agreement(self):
        """

        :return:
        """
        return Agreement(self.template_id, self.conditions[:])

    @property
    def template_id(self):
        """

        :return:
        """
        return self.service_agreement_template.template_id

    @property
    def conditions(self):
        """

        :return:
        """
        return self.service_agreement_template.conditions

    @property
    def condition_by_name(self):
        """

        :return:
        """
        return {cond.name: cond for cond in self.conditions}

    @property
    def conditions_params_value_hashes(self):
        """

        :return:
        """
        value_hashes = []
        for cond in self.conditions:
            value_hashes.append(cond.values_hash)

        return value_hashes

    @property
    def conditions_timeouts(self):
        """

        :return:
        """
        return [cond.timeout for cond in self.conditions]

    @property
    def conditions_timelocks(self):
        """

        :return:
        """
        return [cond.timelock for cond in self.conditions]

    @property
    def conditions_contracts(self):
        """

        :return:
        """
        return [cond.contract_name for cond in self.conditions]

    @classmethod
    def from_ddo(cls, service_type, ddo):
        """

        :param service_type: identifier of the service inside the asset DDO, str
        :param ddo:
        :return:
        """
        service = ddo.get_service(service_type)
        if service is None:
            raise ValueError(
                f'Service of type {service_type} is not found in this DDO.')

        return cls.from_service_dict(service.as_dictionary())

    @classmethod
    def from_service_index(cls, service_index, ddo):
        """

        :param service_index: index of the service inside the asset DDO, str
        :param ddo:
        :return:
        """
        service = ddo.get_service_by_index(service_index)
        if service is None:
            raise ValueError(
                f'Service of type {service_index} is not found in this DDO.')

        return cls.from_service_dict(service.as_dictionary())

    @classmethod
    def from_service_dict(cls, service_dict):
        """

        :param service_dict:
        :return:
        """
        return cls(
            service_dict[cls.SERVICE_ATTRIBUTES],
            ServiceAgreementTemplate(service_dict['templateId'],
                                     service_dict[cls.SERVICE_ATTRIBUTES]['main']['name'],
                                     service_dict[cls.SERVICE_ATTRIBUTES]['main']['creator'],
                                     service_dict[cls.SERVICE_ATTRIBUTES]),
            service_dict.get(cls.SERVICE_ENDPOINT),
            service_dict.get('type')
        )

    @staticmethod
    def generate_service_agreement_hash(template_id, values_hash_list, timelocks, timeouts,
                                        agreement_id, hash_function):
        """

        :param template_id:
        :param values_hash_list:
        :param timelocks:
        :param timeouts:
        :param agreement_id: id of the agreement, hex str
        :param hash_function: reference to function that will be used to compute the hash (sha3
        or similar)
        :return:
        """
        return hash_function(
            ['address', 'bytes32[]', 'uint256[]', 'uint256[]', 'bytes32'],
            [template_id, values_hash_list, timelocks, timeouts, agreement_id]
        )

    @staticmethod
    def create_new_agreement_id():
        """

        :return:
        """
        return generate_prefixed_id()

    def generate_agreement_condition_ids(self, agreement_id, asset_id, consumer_address, keeper, token_address=None):
        """
        Generate the condition ids depending on the ServiceType
        :param agreement_id: id of the agreement, hex str
        :param asset_id:
        :param consumer_address: ethereum account address of consumer, hex str
        :param keeper:
        :param token_address:
        :return:
        """
        if token_address is None:
            token_address = keeper.token.address

        if self.type == ServiceTypes.NFT_ACCESS:
            number_nfts = self.get_number_nfts()
            nft_holder_cond_id = self.generate_nft_holder_condition_id(keeper, agreement_id, asset_id, consumer_address, number_nfts)
            access_cond_id = self.generate_nft_access_condition_id(keeper, agreement_id, asset_id, consumer_address)
            return access_cond_id, nft_holder_cond_id

        amounts = self.get_amounts_int()
        receivers = self.get_receivers()
        lock_cond_id = self.generate_lock_condition_id(keeper, agreement_id, asset_id, keeper.escrow_payment_condition.address, token_address, amounts, receivers)

        if self.type == ServiceTypes.ASSET_ACCESS:
            access_cond_id = self.generate_access_condition_id(keeper, agreement_id, asset_id, consumer_address)

        elif self.type == ServiceTypes.ASSET_ACCESS_PROOF:
            access_cond_id = self.generate_access_proof_condition_id(keeper, agreement_id, asset_id, consumer_address)

        elif self.type == ServiceTypes.CLOUD_COMPUTE:
            access_cond_id = self.generate_compute_condition_id(keeper, agreement_id, asset_id, consumer_address)

        elif self.type == ServiceTypes.DID_SALES:
            access_cond_id = self.generate_transfer_did_condition_id(keeper, agreement_id, asset_id, consumer_address)

        elif self.type == ServiceTypes.NFT_SALES:
            number_nfts = self.get_number_nfts()
            nft_holder = self.get_nft_holder()
            access_cond_id = self.generate_transfer_nft_condition_id(keeper, agreement_id, asset_id, nft_holder, consumer_address, number_nfts, lock_cond_id)

        else:
            raise Exception(
                'Error generating the condition ids, the service_agreement type is not valid.')

        escrow_cond_id = self.generate_escrow_condition_id(keeper, agreement_id, asset_id, keeper.escrow_payment_condition.address, amounts, receivers, token_address, lock_cond_id, access_cond_id)
        return access_cond_id, lock_cond_id, escrow_cond_id

    def generate_nft_holder_condition_id(self, keeper, agreement_id, asset_id, holder_address, number_nfts):
        _hash = add_0x_prefix(
            keeper.nft_holder_condition.hash_values(asset_id, holder_address, number_nfts).hex())
        return add_0x_prefix(
            keeper.nft_holder_condition.contract.functions.generateId(agreement_id, _hash).call().hex())


    def generate_access_condition_id(self, keeper, agreement_id, asset_id, consumer_address):
        _hash = add_0x_prefix(
            keeper.access_condition.hash_values(asset_id, consumer_address).hex())
        return add_0x_prefix(
            keeper.access_condition.contract.functions.generateId(agreement_id, _hash).call().hex())

    def generate_access_proof_condition_id(self, keeper, agreement_id, asset_id, consumer_address):
        keyhash = self.get_key_hash()
        provider_address = self.get_provider_babyjub_key()
        _hash = add_0x_prefix(
            keeper.access_proof_condition.hash_values(keyhash, consumer_address, provider_address).hex())
        return add_0x_prefix(
            keeper.access_proof_condition.contract.functions.generateId(agreement_id, _hash).call().hex())

    def generate_nft_access_condition_id(self, keeper, agreement_id, asset_id, consumer_address):
        _hash = add_0x_prefix(
            keeper.nft_access_condition.hash_values(asset_id, consumer_address).hex())
        return add_0x_prefix(
            keeper.nft_access_condition.contract.functions.generateId(agreement_id, _hash).call().hex())

    def generate_compute_condition_id(self, keeper, agreement_id, asset_id, consumer_address):
        _hash = add_0x_prefix(
            keeper.compute_execution_condition.hash_values(asset_id, consumer_address).hex())
        return add_0x_prefix(
            keeper.compute_execution_condition.contract.functions.generateId(agreement_id, _hash).call().hex())

    def generate_transfer_did_condition_id(self, keeper, agreement_id, asset_id, receiver_address):
        _hash = add_0x_prefix(
            keeper.transfer_did_condition.hash_values(asset_id, receiver_address).hex())
        return add_0x_prefix(
            keeper.transfer_did_condition.contract.functions.generateId(agreement_id, _hash).call().hex())

    def generate_transfer_nft_condition_id(self, keeper, agreement_id, asset_id, nft_holder, receiver_address, number_nfts, lock_cond_id):
        _hash = add_0x_prefix(
            keeper.transfer_nft_condition.hash_values(asset_id, nft_holder, receiver_address, number_nfts, lock_cond_id).hex())
        return add_0x_prefix(
            keeper.transfer_nft_condition.contract.functions.generateId(agreement_id, _hash).call().hex())

    def generate_lock_condition_id(self, keeper, agreement_id, asset_id, escrow_condition_address, token_address, amounts, receivers):
        _hash = add_0x_prefix(
            keeper.lock_payment_condition.hash_values(asset_id, escrow_condition_address, token_address, amounts, receivers).hex())
        return add_0x_prefix(
            keeper.lock_payment_condition.contract.functions.generateId(agreement_id, _hash).call().hex())

    def generate_escrow_condition_id(self, keeper, agreement_id, asset_id, escrow_condition_address, amounts, receivers,token_address, lock_cond_id, access_or_compute_id):
        _hash = add_0x_prefix(
            keeper.escrow_payment_condition.hash_values(asset_id, amounts, receivers, escrow_condition_address, token_address, lock_cond_id, access_or_compute_id).hex())
        return add_0x_prefix(
            keeper.escrow_payment_condition.contract.functions.generateId(agreement_id, _hash).call().hex())

    def get_service_agreement_hash(
            self, agreement_id, asset_id, consumer_address, publisher_address, keeper):
        """Return the hash of the service agreement values to be signed by a consumer.

        :param agreement_id:id of the agreement, hex str
        :param asset_id:
        :param consumer_address: ethereum account address of consumer, hex str
        :param publisher_address: ethereum account address of publisher, hex str
        :param keeper:
        :return:
        """
        agreement_hash = ServiceAgreement.generate_service_agreement_hash(
            self.template_id,
            self.generate_agreement_condition_ids(
                agreement_id, asset_id, consumer_address, keeper),
            self.conditions_timelocks,
            self.conditions_timeouts,
            agreement_id,
            keeper.generate_multi_value_hash
        )
        return agreement_hash



import json
import os
import pathlib
from urllib.request import urlopen

from contracts_lib_py.utils import get_account

from common_utils_py.ddo.ddo import DDO

PUBLISHER_INDEX = 1
CONSUMER_INDEX = 0


def get_resource_path(dir_name, file_name):
    base = os.path.realpath(__file__).split(os.path.sep)[1:-1]
    if dir_name:
        return pathlib.Path(os.path.join(os.path.sep, *base, dir_name, file_name))
    else:
        return pathlib.Path(os.path.join(os.path.sep, *base, file_name))


def get_publisher_account():
    return get_account(0)


def get_consumer_account():
    return get_account(1)


def get_ddo_sample():
    return DDO(json_text=json.dumps(json.loads(urlopen(
        "https://raw.githubusercontent.com/nevermined-io/docs/master/docs/architecture/specs"
        "/examples/access/v0.1/ddo1.json").read().decode(
        'utf-8'))))


def get_ddo_sample2():
    return DDO(json_text=json.dumps(json.loads(urlopen(
        "https://raw.githubusercontent.com/nevermined-io/docs/master/docs/architecture/specs"
        "/examples/access/v0.1/ddo2.json").read().decode(
        'utf-8'))))


def get_ddo_did_sales_sample():
    return DDO(json_text=json.dumps(json.loads(urlopen(
        "https://raw.githubusercontent.com/nevermined-io/docs/master/docs/architecture/specs/examples/access/v0.1"
        "/ddo_sales.json").read().decode(
        'utf-8'))))


def get_ddo_nft_sample():
    return DDO(json_text=json.dumps(json.loads(urlopen(
        "https://raw.githubusercontent.com/nevermined-io/docs/master/docs/architecture/specs/examples/access/v0.1"
        "/ddo_nft.json").read().decode(
        'utf-8'))))


def get_metadata():
    metadata = urlopen(
        "https://raw.githubusercontent.com/nevermined-io/docs/master/docs/architecture/specs"
        "/examples/metadata/v0.1/metadata1.json").read().decode(
        'utf-8')
    return json.loads(metadata)


def log_event(event_name):
    def _process_event(event):
        print(f'Received event {event_name}: {event}')

    return _process_event



"""
    Public key RSA

"""
from .public_key_base import PUBLIC_KEY_STORE_TYPE_BASE64, PUBLIC_KEY_STORE_TYPE_PEM, PublicKeyBase

AUTHENTICATION_TYPE_RSA = 'RsaVerificationKey2018'
PUBLIC_KEY_TYPE_RSA = 'RsaSignatureAuthentication2018'
PUBLIC_KEY_TYPE_ETHEREUM_ECDSA = 'EthereumECDSAKey'


class PublicKeyRSA(PublicKeyBase):
    """Encode key value using RSA."""

    def __init__(self, key_id, **kwargs):
        PublicKeyBase.__init__(self, key_id, **kwargs)
        self._type = PUBLIC_KEY_TYPE_RSA

    def get_authentication_type(self):
        """Return the type of authentication supported by this class."""
        return AUTHENTICATION_TYPE_RSA

    def set_encode_key_value(self, value, store_type=PUBLIC_KEY_STORE_TYPE_BASE64):
        """Set the value based on the type of encoding supported by RSA."""
        if store_type == PUBLIC_KEY_STORE_TYPE_PEM:
            PublicKeyBase.set_encode_key_value(self, value.exportKey('PEM').decode(), store_type)
        else:
            PublicKeyBase.set_encode_key_value(self, value.exportKey('DER'), store_type)



import sqlite3


class StorageBase:
    """
    Provide basic database connection management (connect/close).
    """

    def __init__(self, storage_path):
        self._storage_path = storage_path
        self._conn = None
        if self._storage_path == ':memory:':
            self._conn = sqlite3.connect(self._storage_path)

    def _connect(self):
        if self._storage_path != ':memory:':
            self._conn = sqlite3.connect(self._storage_path)

    def _disconnect(self):
        if self._storage_path != ':memory:':
            self._conn.close()

    def _run_query(self, query, args=None):
        """

        :param query: str the sql query to execute in sqlite3.
        :param args: tuple/list of arguments that go along with the query. Number of arguments
            must match the number of positional `?` in the query string
        :return:
            iterator on rows resulting from the query.
        """
        if not self._conn:
            self._connect()

        cursor = self._conn.cursor()
        result = cursor.execute(query, args or ())
        self._conn.commit()
        return result



"""Helper functions to convert between eth keys and JWK

This only works for EC curve secp256k1.
"""
import base64
from common_utils_py.utils.crypto import get_keys_from_file


def key_bytes_to_jwk(public_key_bytes, private_key_bytes):
    jwk_json = {
        "kty": "EC",
        "crv": "secp256k1",
        "d": base64.urlsafe_b64encode(private_key_bytes),
        "x": base64.urlsafe_b64encode(public_key_bytes[:32]),
        "y": base64.urlsafe_b64encode(public_key_bytes[32:])
    }

    return jwk_json


def key_file_to_jwk(keyfile_path, password):
    public_key_hex, private_key_hex = get_keys_from_file(keyfile_path, password)
    private_key_bytes = bytes.fromhex(private_key_hex[2:])
    public_key_bytes = bytes.fromhex(public_key_hex[2:])
    
    return key_bytes_to_jwk(public_key_bytes, private_key_bytes)


def account_to_jwk(account):
    return key_file_to_jwk(account.key_file, account.password)


from common_utils_py.agreements.access_sla_template import ACCESS_SLA_TEMPLATE
from common_utils_py.agreements.compute_sla_template import COMPUTE_SLA_TEMPLATE
from common_utils_py.agreements.did_sales_template import DID_SALES_TEMPLATE
from common_utils_py.agreements.nft_sales_template import NFT_SALES_TEMPLATE
from common_utils_py.agreements.nft_access_template import NFT_ACCESS_TEMPLATE
from common_utils_py.agreements.access_proof_sla_template import ACCESS_PROOF_SLA_TEMPLATE
from common_utils_py.agreements.service_types import ServiceTypes


def get_sla_template(service_type=ServiceTypes.ASSET_ACCESS):
    """
    Get the template for a ServiceType.

    :param service_type: ServiceTypes
    :return: template dict
    """
    if service_type == ServiceTypes.ASSET_ACCESS:
        return ACCESS_SLA_TEMPLATE.copy()
    elif service_type == ServiceTypes.CLOUD_COMPUTE:
        return COMPUTE_SLA_TEMPLATE.copy()
    elif service_type == ServiceTypes.DID_SALES:
        return DID_SALES_TEMPLATE.copy()
    elif service_type == ServiceTypes.NFT_SALES:
        return NFT_SALES_TEMPLATE.copy()
    elif service_type == ServiceTypes.NFT_ACCESS:
        return NFT_ACCESS_TEMPLATE.copy()
    elif service_type == ServiceTypes.ASSET_ACCESS_PROOF:
        return ACCESS_PROOF_SLA_TEMPLATE.copy()
    else:
        raise ValueError(f'Invalid/unsupported service agreement type {service_type}')



import os

import pytest
from contracts_lib_py.contract_handler import ContractHandler
from contracts_lib_py.keeper import Keeper
from contracts_lib_py.utils import get_account
from contracts_lib_py.web3_provider import Web3Provider
from web3 import HTTPProvider, Web3

from common_utils_py.agreements.service_agreement import ServiceAgreement, ServiceTypes
from common_utils_py.ddo.ddo import DDO
from common_utils_py.did import DID
from common_utils_py.metadata import MetadataProvider
from tests.resources.helper_functions import (get_consumer_account, get_ddo_sample, get_ddo_sample2,
                                              get_metadata, get_publisher_account, get_ddo_did_sales_sample,
                                              get_ddo_nft_sample)
from common_utils_py.utils.utilities import generate_prefixed_id


def get_metadata_url():
    if os.getenv('METADATA_URL'):
        return os.getenv('METADATA_URL')
    return 'http://localhost:5000'


def get_keeper_url():
    if os.getenv('KEEPER_URL'):
        return os.getenv('KEEPER_URL')
    return 'http://localhost:8545'


@pytest.fixture(autouse=True)
def setup_all():
    Web3Provider.get_web3('http://localhost:8545')
    ContractHandler.artifacts_path = os.path.expanduser(
        '~/.nevermined/nevermined-contracts/artifacts')
    Keeper.get_instance()


@pytest.fixture
def publisher_account():
    return get_publisher_account()


@pytest.fixture
def consumer_account():
    return get_consumer_account()


@pytest.fixture
def metadata_instance():
    return MetadataProvider.get_metadata_provider(get_metadata_url())


@pytest.fixture
def registered_ddo():
    return DDO()


@pytest.fixture
def web3_instance():
    return Web3(HTTPProvider(get_keeper_url()))


@pytest.fixture
def metadata():
    return get_metadata()


@pytest.fixture
def json_service():
    return {"type":"nft-sales","index":6,"serviceEndpoint":"https://gateway.rinkeby.nevermined.rocks/api/v1/gateway/services/nft","templateId":"0x24edffc52926739E8403E451b791378349f38818","attributes":{"main":{"name":"nftSalesAgreement","creator":"0xD0064bD1a8DD5a3F775A5432f833EaC9f21CcA80","datePublished":"2021-11-23T10:27:07Z","timeout":86400},"additionalInformation":{"description":""},"serviceAgreementTemplate":{"contractName":"NFTSalesTemplate","events":[{"name":"AgreementCreated","actorType":"consumer","handler":{"moduleName":"nftSalesTemplate","functionName":"fulfillLockPaymentCondition","version":"0.1"}}],"fulfillmentOrder":["lockPayment.fulfill","transferNFT.fulfill","escrowPayment.fulfill"],"conditionDependency":{"lockPayment":[],"transferNFT":[],"escrowPayment":["lockPayment","transferNFT"]},"conditions":[{"name":"lockPayment","timelock":0,"timeout":0,"contractName":"LockPaymentCondition","functionName":"fulfill","parameters":[{"name":"_did","type":"bytes32","value":"688190baee42efb665fb45799135f1511256839e84ccfa7b48616839c49fd427"},{"name":"_rewardAddress","type":"address","value":"0xD0064bD1a8DD5a3F775A5432f833EaC9f21CcA80"},{"name":"_tokenAddress","type":"address","value":"0x937Cc2ec24871eA547F79BE8b47cd88C0958Cc4D"},{"name":"_amounts","type":"uint256[]","value":["20"]},{"name":"_receivers","type":"address[]","value":["0xD0064bD1a8DD5a3F775A5432f833EaC9f21CcA80"]}]}]}}}


@pytest.fixture
def ddo_sample():
    return get_ddo_sample()


@pytest.fixture
def ddo_sample_2():
    return get_ddo_sample2()


@pytest.fixture
def setup_basic_environment():
    consumer_acc = get_consumer_account()
    publisher_acc = get_publisher_account()
    keeper = Keeper.get_instance()
    return (
        keeper,
        publisher_acc,
        consumer_acc
    )

@pytest.fixture
def setup_agreements_environment():
    consumer_acc = get_consumer_account()
    publisher_acc = get_publisher_account()
    keeper = Keeper.get_instance()

    ddo = get_ddo_sample()

    did_seed = generate_prefixed_id()
    asset_id = keeper.did_registry.hash_did(did_seed, publisher_acc.address)

    ddo._did = DID.did(asset_id)

    keeper.did_registry.register(
        did_seed,
        checksum=Web3Provider.get_web3().toBytes(hexstr=ddo.asset_id),
        url='http://172.17.0.1:5000',
        account=publisher_acc,
        providers=None
    )

    service_agreement = ServiceAgreement.from_ddo(ServiceTypes.ASSET_ACCESS, ddo)
    agreement_id = ServiceAgreement.create_new_agreement_id()
    price = service_agreement.get_price()
    (access_cond_id, lock_cond_id, escrow_cond_id) = service_agreement.generate_agreement_condition_ids(
        agreement_id, asset_id, consumer_acc.address, keeper)

    return (
        keeper,
        ddo,
        publisher_acc,
        consumer_acc,
        agreement_id,
        asset_id,
        price,
        service_agreement,
        (lock_cond_id, access_cond_id, escrow_cond_id),
    )


@pytest.fixture
def setup_did_sales_agreements_environment():
    consumer_acc = get_consumer_account()
    publisher_acc = get_publisher_account()
    keeper = Keeper.get_instance()

    ddo = get_ddo_did_sales_sample()

    did_seed = generate_prefixed_id()
    asset_id = keeper.did_registry.hash_did(did_seed, publisher_acc.address)
    ddo._did = DID.did(asset_id)

    keeper.did_registry.register(
        did_seed,
        checksum=Web3Provider.get_web3().toBytes(hexstr=ddo.asset_id),
        url='metadata:5000',
        account=publisher_acc,
        providers=None
    )

    service_agreement = ServiceAgreement.from_ddo(ServiceTypes.DID_SALES, ddo)
    agreement_id = ServiceAgreement.create_new_agreement_id()
    price = service_agreement.get_price()
    (access_cond_id, lock_cond_id, escrow_cond_id) = service_agreement.generate_agreement_condition_ids(
        agreement_id, asset_id, consumer_acc.address, keeper)

    return (
        keeper,
        ddo,
        publisher_acc,
        consumer_acc,
        agreement_id,
        asset_id,
        price,
        service_agreement,
        (lock_cond_id, access_cond_id, escrow_cond_id),
    )


@pytest.fixture
def setup_nft_sales_agreements_environment():
    consumer_acc = get_consumer_account()
    publisher_acc = get_publisher_account()
    keeper = Keeper.get_instance()

    ddo = get_ddo_nft_sample()

    did_seed = generate_prefixed_id()
    asset_id = keeper.did_registry.hash_did(did_seed, publisher_acc.address)
    ddo._did = DID.did(asset_id)

    keeper.did_registry.register_mintable_did(
        did_seed,
        checksum=Web3Provider.get_web3().toBytes(hexstr=ddo.asset_id),
        url='http://172.17.0.1:5000',
        cap=10,
        royalties=10,
        account=publisher_acc,
        providers=None
    )

    keeper.did_registry.mint(ddo.asset_id, 10, account=publisher_acc)

    service_agreement = ServiceAgreement.from_ddo(ServiceTypes.NFT_SALES, ddo)
    agreement_id = ServiceAgreement.create_new_agreement_id()
    price = service_agreement.get_price()
    (access_cond_id, lock_cond_id, escrow_cond_id) = service_agreement.generate_agreement_condition_ids(
        agreement_id, asset_id, consumer_acc.address, keeper)

    nft_access_service_agreement = ServiceAgreement.from_ddo(ServiceTypes.NFT_ACCESS, ddo)
    nft_access_agreement_id = ServiceAgreement.create_new_agreement_id()

    (nft_access_cond_id, nft_holder_cond_id) = nft_access_service_agreement.generate_agreement_condition_ids(
        nft_access_agreement_id, asset_id, consumer_acc.address, keeper)

    return (
        keeper,
        ddo,
        publisher_acc,
        consumer_acc,
        agreement_id,
        nft_access_agreement_id,
        asset_id,
        price,
        service_agreement,
        nft_access_service_agreement,
        (lock_cond_id, access_cond_id, escrow_cond_id),
        (nft_access_cond_id, nft_holder_cond_id)
    )


"""Crypto Utils class"""
import base64

import rsa
from Crypto import Random
from Crypto.Cipher import AES
from Crypto.Protocol.KDF import PBKDF2
from ecies import decrypt, encrypt
from eth_keys import keys
from eth_utils import to_bytes, to_hex
from web3.auto import w3

BLOCK_SIZE = 16
pad = lambda s: s + (BLOCK_SIZE - len(s) % BLOCK_SIZE) * chr(BLOCK_SIZE - len(s) % BLOCK_SIZE)
unpad = lambda s: s[:-ord(s[len(s) - 1:])]


def get_keys_from_file(keyfile_path, keyfile_password):
    with open(keyfile_path) as keyfile:
        encrypted_key = keyfile.read()
    private_key = w3.eth.account.decrypt(encrypted_key, keyfile_password)
    pk = keys.PrivateKey(private_key)

    private_key_hex = to_hex(private_key)  # hex string
    public_key_hex = to_hex(pk.public_key._raw_key)  # hex string

    return public_key_hex, private_key_hex


def get_ecdsa_public_key_from_file(keyfile_path, keyfile_password):
    with open(keyfile_path) as keyfile:
        encrypted_key = keyfile.read()
    private_key = w3.eth.account.decrypt(encrypted_key, keyfile_password)
    pk = keys.PrivateKey(private_key)

    return to_hex(pk.public_key._raw_key)  # hex string


def ecdsa_encryption_from_file(message, provider_key_file, provider_password):
    public_key_hex = get_ecdsa_public_key_from_file(provider_key_file, provider_password)
    encrypted_message = encryption(public_key_hex, message.encode())
    hash = to_hex(encrypted_message)
    return hash, public_key_hex


def ecdsa_decryption(message, provider_key_file, provider_password):
    (public_key_hex, private_key_hex) = get_keys_from_file(provider_key_file, provider_password)
    result = decryption(private_key_hex, to_bytes(hexstr=message))
    return result.decode()


def rsa_encryption_from_file(message, rsa_public_key_file):
    pub_key = get_rsa_public_key_from_file(rsa_public_key_file)
    encrypted_message, aes_encrypted_key = rsa_encryption(pub_key, message.encode())
    hash = to_hex(encrypted_message) + '|' + to_hex(aes_encrypted_key)
    return hash, get_content_keyfile_from_path(rsa_public_key_file)


def get_content_keyfile_from_path(keyfile_path):
    with open(keyfile_path, mode='r') as keyfile:
        key_content = keyfile.read()

    return key_content.replace('-----BEGIN PUBLIC KEY-----', '') \
        .replace('-----END PUBLIC KEY-----', '') \
        .replace('\n', '')


def get_rsa_public_key_from_file(keyfile_path):
    with open(keyfile_path, mode='rb') as keyfile:
        key_content = keyfile.read()
    return rsa.PublicKey.load_pkcs1_openssl_pem(key_content)


def get_rsa_private_key_from_file(keyfile_path):
    with open(keyfile_path, mode='rb') as keyfile:
        key_content = keyfile.read()
    return rsa.PrivateKey.load_pkcs1(key_content)


def rsa_encryption(public_key, data):
    aes_key = rsa.randnum.read_random_bits(128)
    encrypted_data = aes_encryption(data, aes_key)
    encrypted_aes_key = rsa.encrypt(aes_key, public_key)
    return encrypted_data, encrypted_aes_key


def rsa_decryption_aes(message, rsa_private_key_file):
    if '|' in message:  # The message includes an encrypted AES key
        tokens = message.split('|')
        return rsa_decryption(get_rsa_private_key_from_file(rsa_private_key_file),
                              to_bytes(hexstr=tokens[0]), to_bytes(hexstr=tokens[1]))
    priv_key = get_rsa_private_key_from_file(rsa_private_key_file)
    result = rsa.decrypt(to_bytes(hexstr=message), priv_key)
    return result.decode()


def rsa_decryption(private_key, encrypted_data, encrypted_aes_key):
    aes_key = rsa.decrypt(encrypted_aes_key, private_key)
    return aes_decryption(encrypted_data, aes_key)


def get_aes_private_key(passphrase):
    salt = b'this is a salt'
    kdf = PBKDF2(passphrase, salt, 64, 1000)
    key = kdf[:32]
    return key


def aes_encryption(data, passphrase):
    data = data.decode()
    private_key = get_aes_private_key(passphrase)
    data = pad(data).encode()
    iv = Random.new().read(AES.block_size)
    cipher = AES.new(private_key, AES.MODE_CBC, iv)
    return base64.b64encode(iv + cipher.encrypt(data))


def aes_decryption(encrypted_data, passphrase):
    private_key = get_aes_private_key(passphrase)
    encrypted_data = base64.b64decode(encrypted_data)
    iv = encrypted_data[:16]
    cipher = AES.new(private_key, AES.MODE_CBC, iv)
    return unpad(cipher.decrypt(encrypted_data[16:]))


def encryption(public_key_hex, data):
    return encrypt(public_key_hex, data)


def decryption(private_key_hex, encrypted_data):
    return decrypt(private_key_hex, encrypted_data)



"""
    Service Class
    To handle service items in a DDO record
"""
import json
import logging


logger = logging.getLogger(__name__)


class Service:
    """Service class to create validate service in a DDO."""
    SERVICE_ENDPOINT = 'serviceEndpoint'

    def __init__(self, service_endpoint, service_type, values, index=None):
        """Initialize Service instance."""
        self._service_endpoint = service_endpoint
        self._type = service_type
        self._index = index

        # assign the _values property to empty until they are used
        self._values = dict()
        reserved_names = {self.SERVICE_ENDPOINT, 'type'}
        if values:
            for name, value in values.items():
                if name not in reserved_names:
                    self._values[name] = value

    @property
    def type(self):
        """
        Type of the service.

        :return: str
        """
        return self._type

    @property
    def index(self):
        """
        Identifier of the service inside the asset DDO

        :return: str
        """
        return self._index

    @property
    def service_endpoint(self):
        """
        Service endpoint.

        :return: String
        """
        return self._service_endpoint

    def set_service_endpoint(self, service_endpoint):
        """
        Update service endpoint. Needed to update after create did.

        :param service_endpoint: Service endpoint, str
        """
        self._service_endpoint = service_endpoint

    def values(self):
        """

        :return: array of values
        """
        return self._values.copy()

    @property
    def attributes(self):
        return self._values['attributes']

    @property
    def main(self):
        return self._values['attributes']['main']

    def update_value(self, name, value):
        """
        Update value in the array of values.

        :param name: Key of the value, str
        :param value: New value, str
        :return: None
        """
        if name not in {'id', self.SERVICE_ENDPOINT, 'type'}:
            self._values[name] = value

    def as_text(self, is_pretty=False):
        """Return the service as a JSON string."""
        values = {
            'type': self._type,
            self.SERVICE_ENDPOINT: self._service_endpoint,
        }
        if self._values:
            # add extra service values to the dictionary
            for name, value in self._values.items():
                values[name] = value

        if is_pretty:
            return json.dumps(values, indent=4, separators=(',', ': '))

        return json.dumps(values)

    def as_dictionary(self):
        """Return the service as a python dictionary."""
        values = {
            'type': self._type,
            self.SERVICE_ENDPOINT: self._service_endpoint,
        }
        if self._values:
            # add extra service values to the dictionary
            for name, value in self._values.items():
                if isinstance(value, object) and hasattr(value, 'as_dictionary'):
                    value = value.as_dictionary()
                elif isinstance(value, list):
                    value = [v.as_dictionary() if hasattr(v, 'as_dictionary') else v for v in value]

                values[name] = value
        return values

    @classmethod
    def from_json(cls, service_dict):
        """Create a service object from a JSON string."""
        sd = service_dict.copy()
        service_endpoint = sd.get(cls.SERVICE_ENDPOINT)
        if not service_endpoint:
            logger.error(
                'Service definition in DDO document is missing the "serviceEndpoint" key/value.')
            raise IndexError

        _type = sd.get('type')
        _index = sd.get('index')
        if not _type:
            logger.error('Service definition in DDO document is missing the "type" key/value.')
            raise IndexError

        sd.pop(cls.SERVICE_ENDPOINT)
        sd.pop('type')
        return cls(
            service_endpoint,
            _type,
            sd,
            _index
        )



NFT_ACCESS_TEMPLATE = {
  "type": "nft-access",
  "templateId": "",
  "name": "nftAccessAgreement",
  "description": "This service agreement defines the flow for accessing an asset holding a NFT",
  "creator": "",
  "serviceAgreementTemplate": {
    "contractName": "NFTAccessTemplate",
    "events": [
      {
        "name": "AgreementCreated",
        "actorType": "consumer",
        "handler": {
          "moduleName": "nftAccessTemplate",
          "functionName": "fulfillNFTHolderCondition",
          "version": "0.1"
        }
      }
    ],
    "fulfillmentOrder": [
      "nftHolder.fulfill",
      "nftAccess.fulfill"
    ],
    "conditionDependency": {
      "nftHolder": [],
      "access": []
    },
    "conditions": [
      {
        "name": "nftHolder",
        "timelock": 0,
        "timeout": 0,
        "contractName": "NFTHolderCondition",
        "functionName": "fulfill",
        "parameters": [
          {
            "name": "_did",
            "type": "bytes32",
            "value": ""
          },
          {
            "name": "_holderAddress",
            "type": "address",
            "value": ""
          },
          {
            "name": "_numberNfts",
            "type": "uint256",
            "value": ""
          }
        ],
        "events": [
          {
            "name": "Fulfilled",
            "actorType": "publisher",
            "handler": {
              "moduleName": "nftHolderCondition",
              "functionName": "fulfillNFTHolderCondition",
              "version": "0.1"
            }
          }
        ]
      },
      {
        "name": "nftAccess",
        "timelock": 0,
        "timeout": 0,
        "contractName": "NFTAccessCondition",
        "functionName": "fulfill",
        "parameters": [
          {
            "name": "_documentId",
            "type": "bytes32",
            "value": ""
          },
          {
            "name": "_grantee",
            "type": "address",
            "value": ""
          }
        ],
        "events": [
          {
            "name": "Fulfilled",
            "actorType": "publisher",
            "handler": {
              "moduleName": "nftAccess",
              "functionName": "fulfillNFTAccessCondition",
              "version": "0.1"
            }
          },
          {
            "name": "TimedOut",
            "actorType": "consumer",
            "handler": {
              "moduleName": "access",
              "functionName": "fulfillNFTAccessCondition",
              "version": "0.1"
            }
          }
        ]
      }
    ]
  }
}


from common_utils_py.oauth2.jwk_utils import account_to_jwk
from authlib.oauth2.rfc7523.jwt_bearer import JWTBearerGrant
from common_utils_py.utils import keytransfer

BASE_AUD_URL = "/api/v1/gateway/services"


class NeverminedJWTBearerGrant(JWTBearerGrant):
    def create_claims_options(self):
        """Create a claims_options to verify JWT payload claims.
        """
        # https://tools.ietf.org/html/rfc7523#section-3
        claims = {}
        public_claims = {
            'iss': {'essential': True},
            'sub': {
                'essential': False,
                'validate': validate_sub,
            },
            'aud': {
                'essential': True,
                'values': [
                    BASE_AUD_URL + '/access',
                    BASE_AUD_URL + '/access-proof',
                    BASE_AUD_URL + '/nft-access',
                    BASE_AUD_URL + '/compute',
                    BASE_AUD_URL + '/download',
                    BASE_AUD_URL + '/execute'
                ],
            },
            'exp': {'essential': True},
        }
        claims.update(public_claims)

        # private claims are non registered names and may lead to collisions
        private_claims = {
            'did': {
                'essential': False,
                'validate': validate_did
            },
            'execution_id': {
                'essential': False,
                'validate': validate_execution_id
            },
            'eths': {
                'essential': False,
                'values': [
                    "personal"
                ]
            }
        }
        claims.update(private_claims)

        return claims


def validate_sub(claims, value):
    if claims["aud"] == BASE_AUD_URL + "/access" and value is None:
        return False

    return True


def validate_did(claims, value):
    if value is None and claims["aud"] != BASE_AUD_URL + "/compute":
        return False

    return True


def validate_execution_id(claims, value):
    if claims["aud"] != BASE_AUD_URL + "/compute" and value is None:
        return False

    return True


def generate_access_grant_token(account, service_agreement_id, did, uri="/access"):
    # create jwt bearer grant
    jwk = account_to_jwk(account)
    assertion = NeverminedJWTBearerGrant.sign(
        jwk,
        issuer=account.address,
        audience=BASE_AUD_URL + uri,
        subject=service_agreement_id,
        claims={
            "did": did
        },
        header={
            "alg": "ES256K"
        })

    return assertion

def generate_access_proof_grant_token(account, service_agreement_id, did, buyer_secret, uri="/access-proof"):
    # create jwt bearer grant
    jwk = account_to_jwk(account)
    address_num = int(account.address, 16)
    assertion = NeverminedJWTBearerGrant.sign(
        jwk,
        issuer=account.address,
        audience=BASE_AUD_URL + uri,
        subject=service_agreement_id,
        claims={
            "did": did,
            "babysig": keytransfer.sign(buyer_secret, address_num),
            "buyer": keytransfer.make_public(buyer_secret),
        },
        header={
            "alg": "ES256K"
        })

    return assertion


def generate_download_grant_token(account, did):
    # create jwt bearer grant
    jwk = account_to_jwk(account)
    assertion = NeverminedJWTBearerGrant.sign(
        jwk,
        issuer=account.address,
        audience=BASE_AUD_URL + '/download',
        claims={
            "did": did
        },
        header={
            "alg": "ES256K"
        })

    return assertion


def generate_execute_grant_token(account, service_agreement_id, workflow_did):
    # create jwt bearer grant
    jwk = account_to_jwk(account)
    assertion = NeverminedJWTBearerGrant.sign(
        jwk,
        issuer=account.address,
        audience=BASE_AUD_URL + "/execute",
        subject=service_agreement_id,
        claims={
            "did": workflow_did
        },
        header={
            "alg": "ES256K"
        })

    return assertion


def generate_compute_grant_token(account, service_agreement_id, execution_id):
    # create jwt bearer grant
    jwk = account_to_jwk(account)
    assertion = NeverminedJWTBearerGrant.sign(
        jwk,
        issuer=account.address,
        audience=BASE_AUD_URL + "/compute",
        subject=service_agreement_id,
        claims={
            "execution_id": execution_id
        },
        header={
            "alg": "ES256K"
        })

    return assertion



"""
Metadata module.
Help to communicate with the metadata store.
"""
import json
import logging
import copy

from common_utils_py.metadata.exceptions import MetadataGenericError
from common_utils_py.ddo.ddo import DDO
from common_utils_py.ddo.service import Service
from common_utils_py.http_requests.requests_session import get_requests_session

logger = logging.getLogger('metadata')


class Metadata:
    """Metadata wrapper to call different endpoint of metadata component."""

    def __init__(self, metadata_url):
        """
        The Metadata class is a wrapper on the Metadata Store, which has exposed a REST API.

        :param metadata_url: Url of the metadata instance.
        """
        assert metadata_url, f'Invalid url "{metadata_url}"'
        # :HACK:
        if '/api/v1/metadata/assets' in metadata_url:
            metadata_url = metadata_url[:metadata_url.find('/api/v1/metadata/assets')]

        self._base_url = f'{metadata_url}/api/v1/metadata/assets'
        self._headers = {'content-type': 'application/json'}

        logging.debug(f'Metadata Store connected at {metadata_url}')
        logging.debug(f'Metadata Store API documentation at {metadata_url}/api/v1/docs')
        logging.debug(f'Metadata assets at {self._base_url}')

        self.requests_session = get_requests_session()

    @property
    def root_url(self):
        return self._base_url[:self._base_url.find('/api/v1/')]

    @property
    def url(self):
        """Base URL of the metadata instance."""
        return f'{self._base_url}/ddo'

    def get_service_endpoint(self):
        """
        Retrieve the endpoint with the ddo for a given did.

        :return: Return the url of the the ddo location
        """
        return f'{self.url}/' + '{did}'

    def list_assets(self):
        """
        List all the assets registered in the metadata instance.

        :return: List of DID string
        """
        response = self.requests_session.get(self._base_url).content
        if not response:
            return {}

        try:
            asset_list = json.loads(response)
        except TypeError:
            asset_list = None
        except ValueError:
            raise ValueError(response.decode('UTF-8'))

        if not asset_list:
            return []

        if 'ids' in asset_list:
            return asset_list['ids']
        return []

    def get_asset_ddo(self, did):
        """
        Retrieve asset ddo for a given did.

        :param did: Asset DID string
        :return: DDO instance
        """
        response = self.requests_session.get(f'{self.url}/{did}').content
        if not response:
            return {}
        try:
            parsed_response = json.loads(response)
        except TypeError:
            parsed_response = None
        except ValueError:
            raise ValueError(response.decode('UTF-8'))
        if parsed_response is None:
            return {}
        return DDO(dictionary=parsed_response)

    def get_asset_metadata(self, did):
        """
        Retrieve asset metadata for a given did.

        :param did: Asset DID string
        :return: metadata key of the DDO instance
        """
        response = self.requests_session.get(f'{self._base_url}/metadata/{did}').content
        if not response:
            return {}
        try:
            parsed_response = json.loads(response)
        except TypeError:
            parsed_response = None
        except ValueError:
            raise ValueError(response.decode('UTF-8'))
        if parsed_response is None:
            return {}
        return parsed_response['attributes']

    def list_assets_ddo(self):
        """
        List all the ddos registered in the metadata instance.

        :return: List of DDO instance
        """
        return json.loads(self.requests_session.get(self.url).content)

    def publish_asset_ddo(self, ddo):
        """
        Register asset ddo in metadata.

        :param ddo: DDO instance
        :return: API response (depends on implementation)
        """
        try:
            asset_did = ddo.did
            response = self.requests_session.post(self.url, data=ddo.as_text(),
                                                  headers=self._headers)
        except AttributeError:
            raise AttributeError('DDO invalid. Review that all the required parameters are filled.')
        if response.status_code == 500:
            raise ValueError(
                f'This Asset ID already exists! \n\tHTTP Error message: \n\t\t{response.text}')
        elif response.status_code != 201:
            raise Exception(f'{response.status_code} ERROR Full error: \n{response.text}')
        elif response.status_code == 201:
            response = json.loads(response.content)
            logger.debug(f'Published asset DID {asset_did}')
            return response
        else:
            raise Exception(f'Unhandled ERROR: status-code {response.status_code}, '
                            f'error message {response.text}')

    def update_asset_ddo(self, did, ddo):
        """
        Update the ddo of a did already registered.

        :param did: Asset DID string
        :param ddo: DDO instance
        :return: API response (depends on implementation)
        """
        response = self.requests_session.put(f'{self.url}/{did}', data=ddo.as_text(),
                                             headers=self._headers)
        if response.status_code == 200 or response.status_code == 201:
            return json.loads(response.content)
        else:
            raise Exception(f'Unable to update DDO: {response.content}')

    def text_search(self, text, sort=None, offset=100, page=1):
        """
        Search in metadata using text query.

        Given the string metadata will do a full-text query to search in all documents.

        Currently implemented are the MongoDB and Elastic Search drivers.

        For a detailed guide on how to search, see the MongoDB driver documentation:
        mongodb driverCurrently implemented in:
        https://docs.mongodb.com/manual/reference/operator/query/text/

        And the Elastic Search documentation:
        https://www.elastic.co/guide/en/elasticsearch/guide/current/full-text-search.html
        Other drivers are possible according to each implementation.

        :param text: String to be search.
        :param sort: 1/-1 to sort ascending or descending.
        :param offset: Integer with the number of elements displayed per page.
        :param page: Integer with the number of page.
        :return: List of DDO instance
        """
        assert page >= 1, f'Invalid page value {page}. Required page >= 1.'
        payload = {"text": text, "sort": sort, "offset": offset, "page": page}
        response = self.requests_session.get(
            f'{self.url}/query',
            params=payload,
            headers=self._headers
        )
        if response.status_code == 200:
            return self._parse_search_response(response.content)
        else:
            raise Exception(f'Unable to search for DDO: {response.content}')

    def query_search(self, search_query, sort=None, offset=100, page=1):
        """
        Search using a query.

        Currently implemented is the MongoDB query model to search for documents according to:
        https://docs.mongodb.com/manual/tutorial/query-documents/

        And an Elastic Search driver, which implements a basic parser to convert the query into
        elastic search format.

        Example: query_search({"price":[0,10]})

        :param search_query: Python dictionary, query following mongodb syntax
        :param sort: 1/-1 to sort ascending or descending.
        :param offset: Integer with the number of elements displayed per page.
        :param page: Integer with the number of page.
        :return: List of DDO instance
        """
        assert page >= 1, f'Invalid page value {page}. Required page >= 1.'
        search_query['sort'] = sort
        search_query['offset'] = offset
        search_query['page'] = page
        response = self.requests_session.post(
            f'{self.url}/query',
            data=json.dumps(search_query),
            headers=self._headers
        )
        if response.status_code == 200:
            return self._parse_search_response(response.content)
        else:
            raise Exception(f'Unable to search for DDO: {response.content}')

    def retire_asset_ddo(self, did):
        """
        Retire asset ddo of metadata.

        :param did: Asset DID string
        :return: API response (depends on implementation)
        """
        response = self.requests_session.delete(f'{self.url}/{did}', headers=self._headers)
        if response.status_code == 200:
            logging.debug(f'Removed asset DID: {did} from metadata store')
            return response

        raise MetadataGenericError(f'Unable to remove DID: {response}')

    def retire_all_assets(self):
        """
        Retire all the ddo assets.
        :return: str
        """
        response = self.requests_session.delete(f'{self.url}', headers=self._headers)
        if response.status_code == 200:
            logging.debug(f'Removed all the assets successfully')
            return response

        raise MetadataGenericError(f'Unable to remove all the DID: {response}')

    def validate_metadata(self, metadata):
        """
        Validate that the metadata of your ddo is valid.

        :param metadata: conforming to the Metadata accepted by Nevermined, dict
        :return: bool
        """
        response = self.requests_session.post(
            f'{self.url}/validate',
            data=json.dumps(metadata),
            headers=self._headers
        )
        if response.content == b'true\n':
            return True
        else:
            logger.info(self._parse_search_response(response.content))
            return False

    def get_service_agreement(self, agreement_id):
        """
        Retrieve a service agreement for a given agreemend id.

        :param agreement_id: Service agreement id string
        :return: Service
        """
        response = self.requests_session.get(f'{self._base_url}/service/{agreement_id}')
        if response.status_code == 404:
            return None

        if response.status_code == 500:
            raise ValueError(response.content)

        parsed_response = json.loads(response.content)
        return Service.from_json(parsed_response)

    def store_service_agreement(self, agreement_id, service):
        """
        Store a service agreement for a given agreemend id.

        :param agreement_id: Service agreement id string
        :param service: Service
        :return: bool
        """
        service_json = service.as_dictionary()
        service_json['agreementId'] = agreement_id
        response = self.requests_session.post(
            f'{self._base_url}/service',
            data=json.dumps(service_json),
            headers=self._headers
        )

        if response.status_code != 201:
            raise ValueError(response.content)
        return True

    @staticmethod
    def _parse_search_response(response):
        if not response:
            return {}
        try:
            parsed_response = json.loads(response)
        except TypeError:
            parsed_response = None

        if parsed_response is None:
            return []
        elif isinstance(parsed_response, dict):
            return parsed_response
        elif isinstance(parsed_response, list):
            return parsed_response
        else:
            raise ValueError(
                f'Unknown search response, expecting a list got {type(parsed_response)}.')



import logging

from common_utils_py.agreements.service_agreement import ServiceAgreement
from common_utils_py.agreements.service_agreement_template import ServiceAgreementTemplate
from common_utils_py.agreements.service_types import ServiceTypes, ServiceTypesIndices
from common_utils_py.agreements.utils import get_sla_template
from common_utils_py.ddo.service import Service
from common_utils_py.did import did_to_id

logger = logging.getLogger(__name__)

class ServiceDescriptor(object):
    """Tuples of length 2. The first item must be one of ServiceTypes and the second
    item is a dict of parameters and values required by the service"""

    @staticmethod
    def metadata_service_descriptor(attributes, service_endpoint):
        """
        Metadata service descriptor.

        :param attributes: conforming to the Metadata accepted by Nevermined, dict
        :param service_endpoint: identifier of the service inside the asset DDO, str
        :return: Service descriptor.
        """
        return (ServiceTypes.METADATA,
                {'attributes': attributes, 'serviceEndpoint': service_endpoint})

    @staticmethod
    def authorization_service_descriptor(attributes, service_endpoint):
        """
        Authorization service descriptor.

        :param attributes: attributes of the authorization service, dict
        :param service_endpoint: identifier of the service inside the asset DDO, str
        :return: Service descriptor.
        """
        return (ServiceTypes.AUTHORIZATION,
                {'attributes': attributes, 'serviceEndpoint': service_endpoint})

    @staticmethod
    def access_service_descriptor(attributes, service_endpoint):
        """
        Access service descriptor.

        :param attributes: attributes of the access service, dict
        :param service_endpoint: identifier of the service inside the asset DDO, str
        :return: Service descriptor.
        """
        return (
            ServiceTypes.ASSET_ACCESS,
            {'attributes': attributes, 'serviceEndpoint': service_endpoint}
        )

    @staticmethod
    def access_proof_service_descriptor(attributes, service_endpoint):
        """
        Access service with data transfer proof descriptor.

        :param attributes: attributes of the access service, dict
        :param service_endpoint: identifier of the service inside the asset DDO, str
        :return: Service descriptor.
        """
        return (
            ServiceTypes.ASSET_ACCESS_PROOF,
            {'attributes': attributes, 'serviceEndpoint': service_endpoint}
        )

    @staticmethod
    def compute_service_descriptor(attributes, service_endpoint):
        """
        Compute service descriptor.

        :param attributes: attributes of the compute service, dict
        :param service_endpoint: identifier of the service inside the asset DDO, str
        :return: Service descriptor.
        """
        return (
            ServiceTypes.CLOUD_COMPUTE,
            {'attributes': attributes, 'serviceEndpoint': service_endpoint}
        )

    @staticmethod
    def did_sales_service_descriptor(attributes, service_endpoint):
        """
        DID Sales service descriptor.

        :param attributes: attributes of the did sales service, dict
        :param service_endpoint: identifier of the service inside the asset DDO, str
        :return: Service descriptor.
        """
        return (
            ServiceTypes.DID_SALES,
            {'attributes': attributes, 'serviceEndpoint': service_endpoint}
        )

    @staticmethod
    def nft_sales_service_descriptor(attributes, service_endpoint):
        """
        NFT Sales service descriptor.

        :param attributes: attributes of the nft sales service, dict
        :param service_endpoint: identifier of the service inside the asset DDO, str
        :return: Service descriptor.
        """
        return (
            ServiceTypes.NFT_SALES,
            {'attributes': attributes, 'serviceEndpoint': service_endpoint}
        )

    @staticmethod
    def nft_access_service_descriptor(attributes, service_endpoint):
        """
        NFT Access service descriptor.

        :param attributes: attributes of the nft access service, dict
        :param service_endpoint: identifier of the service inside the asset DDO, str
        :return: Service descriptor.
        """
        return (
            ServiceTypes.NFT_ACCESS,
            {'attributes': attributes, 'serviceEndpoint': service_endpoint}
        )


class ServiceFactory(object):
    """Factory class to create Services."""

    @staticmethod
    def build_services(service_descriptors):
        """
        Build a list of services.

        :param service_descriptors: List of tuples of length 2. The first item must be one of
        ServiceTypes
        and the second item is a dict of parameters and values required by the service
        :return: List of Services
        """
        services = []
        for i, service_desc in enumerate(service_descriptors):
            service = ServiceFactory.build_service(service_desc)
            # set index for each service
            service.update_value(ServiceAgreement.SERVICE_INDEX, int(i))
            services.append(service)

        return services

    @staticmethod
    def build_service(service_descriptor):
        """
        Build a service.

        :param service_descriptor: Tuples of length 2. The first item must be one of ServiceTypes
        and the second item is a dict of parameters and values required by the service
        :return: Service
        """
        assert isinstance(service_descriptor, tuple) and len(
            service_descriptor) == 2, 'Unknown service descriptor format.'
        service_type, kwargs = service_descriptor
        if service_type == ServiceTypes.METADATA:
            return ServiceFactory.build_metadata_service(
                kwargs['attributes'],
                kwargs['serviceEndpoint']
            )
        elif service_type == ServiceTypes.AUTHORIZATION:
            return ServiceFactory.build_authorization_service(
                kwargs['attributes'],
                kwargs['serviceEndpoint']
            )
        elif service_type == ServiceTypes.ASSET_ACCESS:
            return ServiceFactory.build_access_service(
                kwargs['attributes'],
                kwargs['serviceEndpoint']
            )
        elif service_type == ServiceTypes.ASSET_ACCESS_PROOF:
            return ServiceFactory.build_access_proof_service(
                kwargs['attributes'],
                kwargs['serviceEndpoint']
            )
        elif service_type == ServiceTypes.CLOUD_COMPUTE:
            return ServiceFactory.build_compute_service(
                kwargs['attributes'],
                kwargs['serviceEndpoint']
            )
        elif service_type == ServiceTypes.DID_SALES:
            return ServiceFactory.build_did_sales_service(
                kwargs['attributes'],
                kwargs['serviceEndpoint']
            )
        elif service_type == ServiceTypes.NFT_SALES:
            return ServiceFactory.build_nft_sales_service(
                kwargs['attributes'],
                kwargs['serviceEndpoint']
            )
        elif service_type == ServiceTypes.NFT_ACCESS:
            return ServiceFactory.build_nft_access_service(
                kwargs['attributes'],
                kwargs['serviceEndpoint']
            )

        raise ValueError(f'Unknown service type {service_type}')

    @staticmethod
    def build_metadata_service(metadata, service_endpoint):
        """
        Build a metadata service.

        :param metadata: conforming to the Metadata accepted by Nevermined, dict
        :param service_endpoint: identifier of the service inside the asset DDO, str
        :return: Service
        """
        return Service(service_endpoint,
                       ServiceTypes.METADATA,
                       values={'attributes': metadata},
                       index=ServiceTypesIndices.DEFAULT_METADATA_INDEX
                       )

    @staticmethod
    def build_authorization_service(attributes, service_endpoint):
        """
        Build an authorization service.

        :param attributes: attributes of authorization service, dict
        :param service_endpoint: identifier of the service inside the asset DDO, str
        :return: Service
        """
        return Service(service_endpoint, ServiceTypes.AUTHORIZATION,
                       values={'attributes': attributes},
                       index=ServiceTypesIndices.DEFAULT_AUTHORIZATION_INDEX)

    @staticmethod
    def build_access_service(attributes, service_endpoint):
        """
        Build an access service.

        :param attributes: attributes of access service, dict
        :param service_endpoint: identifier of the service inside the asset DDO, str
        :return: Service
        """
        return Service(service_endpoint, ServiceTypes.ASSET_ACCESS,
                       values={'attributes': attributes},
                       index=ServiceTypesIndices.DEFAULT_ACCESS_INDEX)

    @staticmethod
    def build_access_proof_service(attributes, service_endpoint):
        """
        Build an access service.

        :param attributes: attributes of access service, dict
        :param service_endpoint: identifier of the service inside the asset DDO, str
        :return: Service
        """
        return Service(service_endpoint, ServiceTypes.ASSET_ACCESS_PROOF,
                       values={'attributes': attributes},
                       index=ServiceTypesIndices.DEFAULT_ACCESS_PROOF_INDEX)

    @staticmethod
    def build_compute_service(attributes, service_endpoint):
        """
        Build a compute service.

        :param attributes: attributes of compute service, dict
        :param service_endpoint: identifier of the service inside the asset DDO, str
        :return: Service
        """
        return Service(service_endpoint, ServiceTypes.CLOUD_COMPUTE,
                       values={'attributes': attributes},
                       index=ServiceTypesIndices.DEFAULT_COMPUTING_INDEX)

    @staticmethod
    def build_did_sales_service(attributes, service_endpoint):
        """
        Build a did sales service.

        :param attributes: attributes of did sales service, dict
        :param service_endpoint: identifier of the service inside the asset DDO, str
        :return: Service
        """
        return Service(service_endpoint, ServiceTypes.DID_SALES,
                       values={'attributes': attributes},
                       index=ServiceTypesIndices.DEFAULT_DID_SALES_INDEX)

    @staticmethod
    def build_nft_sales_service(attributes, service_endpoint):
        """
        Build a nft sales service.

        :param attributes: attributes of nft sales service, dict
        :param service_endpoint: identifier of the service inside the asset DDO, str
        :return: Service
        """
        return Service(service_endpoint, ServiceTypes.NFT_SALES,
                       values={'attributes': attributes},
                       index=ServiceTypesIndices.DEFAULT_NFT_SALES_INDEX)

    @staticmethod
    def build_nft_access_service(attributes, service_endpoint):
        """
        Build a nft sales service.

        :param attributes: attributes of nft sales service, dict
        :param service_endpoint: identifier of the service inside the asset DDO, str
        :return: Service
        """
        return Service(service_endpoint, ServiceTypes.NFT_ACCESS,
                       values={'attributes': attributes},
                       index=ServiceTypesIndices.DEFAULT_NFT_ACCESS_INDEX)

    @staticmethod
    def complete_access_service(did, service_endpoint, attributes, template_id,
                                reward_contract_address=None, service_type=ServiceTypes.ASSET_ACCESS):
        """
        Build the access service.

        :param did: DID, str
        :param service_endpoint: identifier of the service inside the asset DDO, str
        :param template_id: id of the template use to create the service, str
        :param reward_contract_address: hex str ethereum address of deployed reward condition
            smart contract
        :return: ServiceAgreement
        """
        param_map = {
            '_documentId': did_to_id(did),
            '_amount': attributes['main']['price']
        }

        if reward_contract_address is not None:
            param_map ['_rewardAddress'] = reward_contract_address

        try:
            param_map['_did'] = did_to_id(did)
            param_map['_amounts'] = attributes['main']['_amounts']
            param_map['_receivers'] = attributes['main']['_receivers']
            param_map['_tokenAddress'] = attributes['main']['_tokenAddress']
            param_map['_numberNfts'] = attributes['main']['_numberNfts']
        except KeyError as e:
            logger.error(f'Error mapping field {e}')

        sla_template_dict = get_sla_template(service_type)
        sla_template = ServiceAgreementTemplate(template_id, service_type,
                                                attributes['main']['creator'], sla_template_dict)
        sla_template.template_id = template_id
        conditions = sla_template.conditions[:]

        for cond in conditions:
            for param in cond.parameters:
                param.value = param_map.get(param.name, '')

            if cond.timeout > 0:
                cond.timeout = attributes['main']['timeout']

        sla_template.set_conditions(conditions)
        sa = ServiceAgreement(
            attributes,
            sla_template,
            service_endpoint,
            service_type
        )
        return sa

    @staticmethod
    def complete_compute_service(did, service_endpoint, attributes, template_id,
                                 reward_contract_address):
        """
        Build the access service.

        :param did: DID, str
        :param service_endpoint: identifier of the service inside the asset DDO, str
        :param template_id: id of the template use to create the service, str
        :param reward_contract_address: hex str ethereum address of deployed reward condition
            smart contract
        :return: ServiceAgreement
        """
        param_map = {
            '_documentId': did_to_id(did),
            '_amount': attributes['main']['price'],
            '_rewardAddress': reward_contract_address
        }

        try:
            param_map['_amounts'] = attributes['main']['_amounts']
            param_map['_receivers'] = attributes['main']['_receivers']
        except KeyError:
            pass

        sla_template_dict = get_sla_template(ServiceTypes.CLOUD_COMPUTE)
        sla_template = ServiceAgreementTemplate(template_id, ServiceTypes.CLOUD_COMPUTE,
                                                attributes['main']['creator'], sla_template_dict)
        sla_template.template_id = template_id
        conditions = sla_template.conditions[:]
        for cond in conditions:
            for param in cond.parameters:
                param.value = param_map.get(param.name, '')

            if cond.timeout > 0:
                cond.timeout = attributes['main']['timeout']

        sla_template.set_conditions(conditions)
        sa = ServiceAgreement(
            attributes,
            sla_template,
            service_endpoint,
            ServiceTypes.CLOUD_COMPUTE
        )
        return sa

    @staticmethod
    def complete_nft_sales_service(did, service_endpoint, attributes, template_id,
                                reward_contract_address=None, service_type=ServiceTypes.NFT_SALES):
        """
        Build the nft sales service.

        :param did: DID, str
        :param service_endpoint: identifier of the service inside the asset DDO, str
        :param template_id: id of the template use to create the service, str
        :param reward_contract_address: hex str ethereum address of deployed reward condition
            smart contract
        :return: ServiceAgreement
        """
        param_map = {
            '_documentId': did_to_id(did),
            '_amount': attributes['main']['price']
        }

        if reward_contract_address is not None:
            param_map ['_rewardAddress'] = reward_contract_address

        try:
            param_map['_amounts'] = attributes['main']['_amounts']
            param_map['_receivers'] = attributes['main']['_receivers']
            param_map['_numberNfts'] = attributes['main']['_numberNfts']
            param_map['_nftHolder'] = attributes['main']['_nftHolder']
        except KeyError:
            pass

        sla_template_dict = get_sla_template(service_type)
        sla_template = ServiceAgreementTemplate(template_id, service_type,
                                                attributes['main']['creator'], sla_template_dict)
        sla_template.template_id = template_id
        conditions = sla_template.conditions[:]

        for cond in conditions:
            for param in cond.parameters:
                param.value = param_map.get(param.name, '')

            if cond.timeout > 0:
                cond.timeout = attributes['main']['timeout']

        sla_template.set_conditions(conditions)
        sa = ServiceAgreement(
            attributes,
            sla_template,
            service_endpoint,
            service_type
        )
        return sa




from web3 import Web3

from common_utils_py.utils import utilities


def test_convert():
    input_text = "my text"
    text_bytes = utilities.convert_to_bytes(Web3, input_text)
    print("output %s" % utilities.convert_to_string(Web3, text_bytes))
    assert input_text == utilities.convert_to_text(Web3, text_bytes)



import json
import logging

from eth_utils import add_0x_prefix

from common_utils_py.agreements.service_agreement import ServiceAgreement
from common_utils_py.agreements.service_types import ServiceTypes
from common_utils_py.ddo.public_key_base import PublicKeyBase
from common_utils_py.ddo.public_key_rsa import PUBLIC_KEY_TYPE_ETHEREUM_ECDSA
from common_utils_py.did import did_to_id, NEVERMINED_PREFIX
from common_utils_py.utils.utilities import get_timestamp
from .constants import DID_DDO_CONTEXT_URL, PROOF_TYPE
from .public_key_rsa import PUBLIC_KEY_TYPE_RSA, PublicKeyRSA
from .service import Service

logger = logging.getLogger('ddo')


class DDO:
    """DDO class to create, import, export, validate DDO objects."""

    def __init__(self, did=None, json_text=None, json_filename=None, created=None, dictionary=None):
        """Clear the DDO data values."""
        self._did = did
        self._public_keys = []
        self._authentications = []
        self._services = []
        self._proof = None
        self._created = None

        if created:
            self._created = created
        else:
            self._created = get_timestamp()

        if not json_text and json_filename:
            with open(json_filename, 'r') as file_handle:
                json_text = file_handle.read()

        if json_text:
            self._read_dict(json.loads(json_text))
        elif dictionary:
            self._read_dict(dictionary)

    @property
    def did(self):
        """ Get the DID."""
        return self._did

    @property
    def asset_id(self):
        """The asset id part of the DID"""
        if not self._did:
            return None
        return add_0x_prefix(did_to_id(self._did))

    @property
    def services(self):
        """Get the list of services."""
        return self._services[:]

    @property
    def proof(self):
        """Get the static proof, or None."""
        return self._proof

    @property
    def publisher(self):
        return self._proof.get('creator') if self._proof else None

    @property
    def metadata(self):
        """Get the metadata service."""
        metadata_service = self.get_service(ServiceTypes.METADATA)
        return metadata_service.attributes if metadata_service else None

    @property
    def authorization(self):
        """Get the metadata service."""
        authorization_service = self.get_service(ServiceTypes.AUTHORIZATION)
        return authorization_service

    @property
    def created(self):
        return self._created

    @property
    def encrypted_files(self):
        """Return encryptedFiles field in the base metadata."""
        files = self.metadata['encryptedFiles']
        return files

    def assign_did(self, did):
        if self._did:
            raise AssertionError('"did" is already set on this DDO instance.')
        assert did and isinstance(did, str), \
            f'did must be of str type, got {did} of type {type(did)}'
        assert did.startswith(NEVERMINED_PREFIX), \
            f'"did" seems invalid, must start with {NEVERMINED_PREFIX} prefix.'
        self._did = did
        return did

    def add_public_key(self, did, public_key):
        """
        Add a public key object to the list of public keys.

        :param public_key: Public key, PublicKeyHex
        """
        logger.debug(f'Adding public key {public_key} to the did {did}')
        self._public_keys.append(
            PublicKeyBase(did, **{"owner": public_key, "type": PUBLIC_KEY_TYPE_ETHEREUM_ECDSA}))

    def add_authentication(self, public_key, authentication_type):
        """
        Add a authentication public key id and type to the list of authentications.

        :param public_key: Key id, Authentication
        :param authentication_type: Authentication type, str
        """
        authentication = {}
        if public_key:
            authentication = {'type': authentication_type, 'publicKey': public_key}
        logger.debug(f'Adding authentication {authentication}')
        self._authentications.append(authentication)

    def add_service(self, service_type, service_endpoint=None, values=None, index=None):
        """
        Add a service to the list of services on the DDO.

        :param service_type: Service
        :param service_endpoint: Service endpoint, str
        :param values: Python dict with index, templateId, serviceAgreementContract,
        list of conditions and purchase endpoint.
        """
        if isinstance(service_type, Service):
            service = service_type
        else:
            service = Service(service_endpoint, service_type, values, index)
        logger.debug(f'Adding service with service type {service_type} with did {self._did}')
        self._services.append(service)

    def as_text(self, is_proof=True, is_pretty=False):
        """Return the DDO as a JSON text.

        :param if is_proof: if False then do not include the 'proof' element.
        :param is_pretty: If True return dictionary in a prettier way, bool
        :return: str
        """
        data = self.as_dictionary(is_proof)
        if is_pretty:
            return json.dumps(data, indent=2, separators=(',', ': '))

        return json.dumps(data)

    def as_dictionary(self, is_proof=True):
        """
        Return the DDO as a JSON dict.

        :param if is_proof: if False then do not include the 'proof' element.
        :return: dict
        """
        if self._created is None:
            self._created = get_timestamp()

        data = {
            '@context': DID_DDO_CONTEXT_URL,
            'id': self._did,
            'created': self._created,
        }
        if self._public_keys:
            values = []
            for public_key in self._public_keys:
                values.append(public_key.as_dictionary())
            data['publicKey'] = values
        if self._authentications:
            values = []
            for authentication in self._authentications:
                values.append(authentication)
            data['authentication'] = values
        if self._services:
            values = []
            for service in self._services:
                values.append(service.as_dictionary())
            data['service'] = values
        if self._proof and is_proof:
            data['proof'] = self._proof

        return data

    def _read_dict(self, dictionary):
        """Import a JSON dict into this DDO."""
        values = dictionary
        self._did = values['id']
        self._created = values.get('created', None)
        if 'publicKey' in values:
            self._public_keys = []
            for value in values['publicKey']:
                if isinstance(value, str):
                    value = json.loads(value)
                self._public_keys.append(DDO.create_public_key_from_json(value))
        if 'authentication' in values:
            self._authentications = []
            for value in values['authentication']:
                if isinstance(value, str):
                    value = json.loads(value)
                self._authentications.append(DDO.create_authentication_from_json(value))
        if 'service' in values:
            self._services = []
            for value in values['service']:
                if isinstance(value, str):
                    value = json.loads(value)
                if value['type'] == ServiceTypes.ASSET_ACCESS:
                    service = ServiceAgreement.from_service_dict(value)
                elif value['type'] == ServiceTypes.ASSET_ACCESS_PROOF:
                    service = ServiceAgreement.from_service_dict(value)
                elif value['type'] == ServiceTypes.CLOUD_COMPUTE:
                    service = ServiceAgreement.from_service_dict(value)
                elif value['type'] == ServiceTypes.NFT_SALES:
                    service = ServiceAgreement.from_service_dict(value)
                else:
                    service = Service.from_json(value)
                self._services.append(service)
        if 'proof' in values:
            self._proof = values['proof']

    def add_proof(self, checksums, publisher_account):
        """Add a proof to the DDO, based on the public_key id/index and signed with the private key
        add a static proof to the DDO, based on one of the public keys.

        :param checksums: dict with the checksum of the main attributes of each service, dict
        :param publisher_account: account of the publisher, account
        """
        self._proof = {
            'type': PROOF_TYPE,
            'created': get_timestamp(),
            'creator': publisher_account.address,
            'signatureValue': '',
            'checksum': checksums

        }

    def get_public_key(self, key_id):
        """Key_id can be a string, or int. If int then the index in the list of keys."""
        if isinstance(key_id, int):
            return self._public_keys[key_id]

        for item in self._public_keys:
            if item.get_id() == key_id:
                return item

        return None

    def _get_public_key_count(self):
        """Return the count of public keys in the list and embedded."""
        return len(self._public_keys)

    def _get_authentication_from_public_key_id(self, key_id):
        """Return the authentication based on it's id."""
        for authentication in self._authentications:
            if authentication['publicKey'] == key_id:
                return authentication
        return None

    def get_service(self, service_type=None):
        """Return a service using."""
        for service in self._services:
            if service.type == service_type and service_type:
                return service
        return None

    def get_service_by_index(self, index):
        """
        Get service for a given index.

        :param index: Service id, str
        :return: Service
        """
        try:
            index = int(index)
        except ValueError:
            logging.error(f'The index {index} can not be converted into a int')
            return None

        for service in self._services:
            if service.index == index:
                return service

        # try to find by type
        return self.get_service(index)

    @property
    def public_keys(self):
        """Get the list of public keys."""
        return self._public_keys[:]

    @property
    def authentications(self):
        """Get the list authentication records."""
        return self._authentications[:]

    @staticmethod
    def create_public_key_from_json(values):
        """Create a public key object based on the values from the JSON record."""
        # currently we only support RSA public keys
        _id = values.get('id')
        if not _id:
            # Make it more forgiving for now.
            _id = ''
            # raise ValueError('publicKey definition is missing the "id" value.')

        if values.get('type') == PUBLIC_KEY_TYPE_RSA:
            public_key = PublicKeyRSA(_id, owner=values.get('owner'))
        else:
            public_key = PublicKeyBase(
                _id, owner=values.get('owner'), type=PUBLIC_KEY_TYPE_ETHEREUM_ECDSA)

        public_key.set_key_value(values)
        return public_key

    @staticmethod
    def create_authentication_from_json(values):
        """Create authentication object from a JSON dict."""
        key_id = values.get('publicKey')
        authentication_type = values.get('type')
        if not key_id:
            raise ValueError(
                f'Invalid authentication definition, "publicKey" is missing: {values}')

        authentication = {'type': authentication_type, 'publicKey': key_id}
        return authentication



from common_utils_py.utils.poseidon_constants import constants
from common_utils_py.utils.mimc_constants import mimc_constants
from ctypes import *
import json
from web3 import Web3

F = 21888242871839275222246405745257275088548364400416034343698204186575808495617

N_ROUNDS_F = 8
N_ROUNDS_P = [56, 57, 56, 60, 60, 63, 64, 63, 60, 66, 60, 65, 70, 60, 64, 68]

def C(a,b):
    return int(constants['C'][a][b], 16)

def M(a,b,c):
    return int(constants['M'][a][b][c], 16)

def square(a):
    return (a ** 2) % F

def pow5(a):
    return (a ** 5) % F

def add(a,b):
    return (a+b) % F

def sub(a,b):
    return (a-b + F) % F

def div(a,b):
    return mul(a, pow(b, F-2, F))
#    return mul(a, pow(b, -1, F))

def mul(a,b):
    return (a*b) % F

def poseidon(inputs):
    assert(len(inputs) > 0)
    assert(len(inputs) < len(N_ROUNDS_P) - 1)

    t = len(inputs) + 1
    nRoundsF = N_ROUNDS_F
    nRoundsP = N_ROUNDS_P[t - 2]

    state = [0] + inputs
    for r in range(0,nRoundsF + nRoundsP):
        nstate = []
        for i in range(0, len(state)):
            nstate.append(add(state[i], C(t - 2,r * t + i)))
        state = nstate

        if (r < nRoundsF / 2 or r >= nRoundsF / 2 + nRoundsP):
            state = list(map(pow5, state))
        else:
            state[0] = pow5(state[0])

        nstate = []
        for i in range(0, len(state)):
            acc = 0
            for j in range(0, len(state)):
                acc = add(acc, mul(M(t - 2,i,j), state[j]))
            nstate.append(acc)
        state = nstate
    return state[0] % F

NROUNDS = 220

def cts(a):
    return int(mimc_constants[a], 16)

def mimc(_xL_in, _xR_in, _k):
    xL = _xL_in
    xR = _xR_in
    k = _k
    for i in range(0, NROUNDS):
        c = cts(i)
        if i==0:
            t = add(xL, k)
        else:
            t = add(add(xL, k), c)
        xR_tmp = xR
        if i < (NROUNDS - 1):
            xR = xL
            xL = add(xR_tmp, pow5(t))
        else:
            xR = add(xR_tmp, pow5(t))
    return [xL%F, xR%F]

generator = [
    995203441582195749578291179787384436505546430278305826713579947235728471134,
    5472060717959818805561601436314318772137091100104008585924551046643952123905,
]
base8 = [
    5299619240641551281634865583518297030282874472190772894086521144482721001553,
    16950150798460657717958625567821834550301663161624707787222815936182638968203,
]
order = 21888242871839275222246405745257275088614511777268538073601725287587578984328
subOrder = order // 8
A = 168700
D = 168696

def addPoint(a,b):
    beta = mul(a[0],b[1])
    gamma = mul(a[1],b[0])
    delta = mul(sub(a[1], mul(A, a[0])), add(b[0], b[1]))
    tau = mul(beta, gamma)
    dtau = mul(D, tau)
    return [div(add(beta, gamma), add(1, dtau)), div(add(delta, sub(mul(A,beta), gamma)), sub(1, dtau))]

def mulPointEscalar(base, e):
    res = [0,1]
    rem = e
    exp = base

    while rem != 0:
        if rem % 2 == 1:
            res = addPoint(res, exp)
        exp = addPoint(exp, exp)
        rem = rem // 2

    return res

def inCurve(P):
    x2 = square(P[0])
    y2 = square(P[1])

    if add(mul(A, x2), y2) != add(1, mul(mul(x2, y2), D)):
        return False

    return True

def make_key(provider_secret):
    c = Web3.keccak(text=provider_secret)
    return int(c.hex()[0:60], 16)

def make_public(secret):
    k = make_key(secret)
    buyerPub = mulPointEscalar(base8, k)
    res = hex(buyerPub[0])[2:].zfill(64) + hex(buyerPub[1])[2:].zfill(64)
    return res

def sign(provider_secret, msg):
    r = make_key(provider_secret + 'a')
    r = r % subOrder
    s = make_key(provider_secret)
    R8 = mulPointEscalar(base8, r)
    A = mulPointEscalar(base8, s)
    hm = poseidon([R8[0], R8[1], A[0], A[1], msg % F])
    S = (r + hm*s) % subOrder
    return {
        'R8': [hex(R8[0]), hex(R8[1])],
        'S': hex(S),
    }

def verify(A, msg, sig):
    sig_R8 = [int(sig['R8'][0], 16), int(sig['R8'][1], 16)]
    sig_S = int(sig['S'], 16)
    if not inCurve(sig_R8):
        return False
    if not inCurve(A):
        return False
    if sig_S >= subOrder:
        return False

    hm = poseidon([sig_R8[0], sig_R8[1], A[0], A[1], msg % F])

    Pleft = mulPointEscalar(base8, sig_S*8)
    Pright = mulPointEscalar(A, hm*8)
    Pright = addPoint(mulPointEscalar(sig_R8, 8), Pright)

    if Pleft[0] != Pright[0]:
        return False
    if Pleft[1] != Pright[1]:
        return False
    return True

libkey = 0

def init_prover():
    global libkey
    libname = "libkeytransfer.so"
    file = open("/proc/cpuinfo").read()
    if not 'adx' in file:
        libname = "libkeytransfer_noadx.so"
    cdll.LoadLibrary(libname)
    libkey = CDLL(libname)
    libkey.make.restype = c_void_p
    libkey.fullprove.restype = c_char_p

def make_prover(zkey, dat):
    if libkey == 0:
        init_prover()
    return libkey.make(zkey.encode('utf-8'), dat.encode('utf-8'))

def prove(prover, input):
    with open('/tmp/input.json', 'w') as outfile:
        json.dump(input, outfile)
    res = libkey.fullprove(c_void_p(prover), b"/tmp/keytransfer.wtns", b"/tmp/input.json")
    return res.decode()

def split(data):
    return [int(data[0:16].hex(), 16), int(data[16:32].hex(), 16)]

def hash_key(data):
    lst = split(data)
    return hex(poseidon(lst))

def hx(a):
    return str(a)

def prove_transfer(prover, buyerPub, providerK, data):
    orig = split(data)

    k = mulPointEscalar(buyerPub, providerK)
    cipher = mimc(orig[0], orig[1], k[0])
    origHash = poseidon([orig[0], orig[1]])
    providerPub = mulPointEscalar(base8, providerK)


    snarkParams = {
        'buyer_x': hx(buyerPub[0]),
        'buyer_y': hx(buyerPub[1]),
        'provider_x': hx(providerPub[0]),
        'provider_y': hx(providerPub[1]),
        'xL_in': hx(orig[0]),
        'xR_in': hx(orig[1]),
        'cipher_xL_in': hx(cipher[0]),
        'cipher_xR_in': hx(cipher[1]),
        'provider_k': hx(providerK),
        'hash_plain': hx(origHash)
    }

    res = {
        'proof': prove(prover, snarkParams),
        'hash': hash_key(data),
        'cipher': [hex(cipher[0]), hex(cipher[1])]
    }
    return res




class AdditionalInfoMeta(object):
    """Attributes that can enhance the discoverability of a resource."""
    KEY = 'additionalInformation'
    VALUES_KEYS = (
        'categories',
        'copyrightHolder',
        'description',
        'inLanguage',
        'links',
        'tags',
        'updateFrequency',
        'structuredMarkup',
        'poseidonHash',
        'providerKey',
        'workExample'
    )
    REQUIRED_VALUES_KEYS = tuple()


class CurationMeta(object):
    """To normalize the different possible rating attributes after a curation process."""
    KEY = 'curation'
    VALUES_KEYS = (
        "rating", "numVotes", "schema", 'isListed'
    )
    REQUIRED_VALUES_KEYS = {'rating', 'numVotes'}


class MetadataMain(object):
    """The main attributes that need to be included in the Asset Metadata."""
    KEY = 'main'
    VALUES_KEYS = {
        'author',
        'name',
        'type',
        'dateCreated',
        'license',
        'price',
        'files'
    }
    REQUIRED_VALUES_KEYS = {'name', 'dateCreated', 'author', 'license', 'price', 'files'}


class Metadata(object):
    """Every Asset (dataset, algorithm, etc.) in a Nevermined Network has an associated Decentralized
    Identifier (DID) and DID document / DID Descriptor Object (DDO)."""
    REQUIRED_SECTIONS = {MetadataMain.KEY}
    MAIN_SECTIONS = {
        MetadataMain.KEY: MetadataMain,
        CurationMeta.KEY: CurationMeta,
        AdditionalInfoMeta.KEY: AdditionalInfoMeta
    }

    @staticmethod
    def validate(metadata):
        """Validator of the metadata composition

        :param metadata: conforming to the Metadata accepted by Nevermined, dict
        :return: bool
        """
        # validate required sections and their sub items
        for section_key in Metadata.REQUIRED_SECTIONS:
            if section_key not in metadata or not metadata[section_key] or not isinstance(
                    metadata[section_key], dict):
                return False

            section = Metadata.MAIN_SECTIONS[section_key]
            section_metadata = metadata[section_key]
            for subkey in section.REQUIRED_VALUES_KEYS:
                if subkey not in section_metadata or section_metadata[subkey] is None:
                    return False

        return True



"""Exceptions for common_utils_py."""


class InvalidContractAddress(Exception):
    """Raised when an invalid address is passed to the contract loader."""


class DIDUnknownValueType(Exception):
    """Raised when a requested DID or a DID in the chain cannot be found."""


class DIDAlreadyExist(Exception):
    """Raised when a requested DID is already published in Metadata API."""


class InvalidMetadata(Exception):
    """Raised when some value in the metadata is invalid."""


class InvalidServiceAgreementSignature(Exception):
    """Raised when the SLA signature is not valid."""


class ServiceAgreementExists(Exception):
    """Raised when the SLA already exists."""


class InitializeServiceAgreementError(Exception):
    """Error on invoking purchase endpoint"""


class EncryptAssetUrlsError(Exception):
    """Error invoking the encrypt endpoint"""


class ServiceConsumeError(Exception):
    """ Error invoking a purchase endpoint"""


class InvalidAgreementTemplate(Exception):
    """ Error when agreement template is not valid or not approved"""



class ServiceTypes:
    """Types of Service allowed in ocean protocol DDO services."""
    AUTHORIZATION = 'authorization'
    METADATA = 'metadata'
    ASSET_ACCESS = 'access'
    ASSET_ACCESS_PROOF = 'access-proof'
    CLOUD_COMPUTE = 'compute'
    DID_SALES = 'did-sales'
    NFT_SALES = 'nft-sales'
    NFT_ACCESS = 'nft-access'
    NFT721_ACCESS = 'nft721-access'


class ServiceTypesIndices:
    DEFAULT_METADATA_INDEX = 0
    DEFAULT_PROVENANCE_INDEX = 1
    DEFAULT_AUTHORIZATION_INDEX = 2
    DEFAULT_ACCESS_INDEX = 3
    DEFAULT_COMPUTING_INDEX = 4
    DEFAULT_DID_SALES_INDEX = 5
    DEFAULT_NFT_SALES_INDEX = 6
    DEFAULT_NFT_ACCESS_INDEX = 7
    DEFAULT_NFT721_ACCESS_INDEX = 9
    DEFAULT_ACCESS_PROOF_INDEX = 10


class ServiceAuthorizationTypes:
    SECRET_STORE = 'SecretStore'
    PSK_ECDSA = 'PSK-ECDSA'
    PSK_RSA = 'PSK-RSA'



COMPUTE_SLA_TEMPLATE = {
  "type": "Compute",
  "templateId": "",
  "name": "dataAssetComputeExecutionAgreement",
  "description": "",
  "creator": "",
  "serviceAgreementTemplate": {
    "contractName": "EscrowComputeExecutionTemplate",
    "events": [
      {
        "name": "AgreementCreated",
        "actorType": "consumer",
        "handler": {
          "moduleName": "EscrowComputeExecutionTemplate",
          "functionName": "fulfillLockPaymentCondition",
          "version": "0.1"
        }
      }
    ],
    "fulfillmentOrder": [
      "lockPayment.fulfill",
      "execCompute.fulfill",
      "escrowPayment.fulfill"
    ],
    "conditionDependency": {
      "lockPayment": [],
      "execCompute": [],
      "escrowPayment": [
        "lockPayment",
        "execCompute"
      ]
    },
    "conditions": [
        {
            "name": "lockPayment",
            "timelock": 0,
            "timeout": 0,
            "contractName": "LockPaymentCondition",
            "functionName": "fulfill",
            "parameters": [
                {
                    "name": "_did",
                    "type": "bytes32",
                    "value": "{parameter.assetId}"
                },
                {
                    "name": "_rewardAddress",
                    "type": "address",
                    "value": "{contract.EscrowPayment.address}"
                },
                {
                    "name": "_tokenAddress",
                    "type": "address",
                    "value": ""
                },
                {
                    "name": "_amounts",
                    "type": "uint256[]",
                    "value": []
                },
                {
                    "name": "_receivers",
                    "type": "address[]",
                    "value": []
                }
            ],
            "events": [
                {
                    "name": "Fulfilled",
                    "actorType": "publisher",
                    "handler": {
                        "moduleName": "lockPaymentCondition",
                        "functionName": "fulfillExecComputeCondition",
                        "version": "0.1"
                    }
                }
            ]
        },
        {
            "name": "execCompute",
            "timelock": 0,
            "timeout": 0,
            "contractName": "ComputeExecutionCondition",
            "functionName": "fulfill",
            "parameters": [
                {
                    "name": "_documentId",
                    "type": "bytes32",
                    "value": "{parameter.assetId}"
                },
                {
                    "name": "_grantee",
                    "type": "address",
                    "value": ""
                }
            ],
            "events": [
                {
                    "name": "Fulfilled",
                    "actorType": "publisher",
                    "handler": {
                        "moduleName": "access",
                        "functionName": "fulfillEscrowPaymentCondition",
                        "version": "0.1"
                    }
                },
                {
                    "name": "TimedOut",
                    "actorType": "consumer",
                    "handler": {
                        "moduleName": "execCompute",
                        "functionName": "fulfillEscrowPaymentCondition",
                        "version": "0.1"
                    }
                }
            ]
        },
        {
            "name": "escrowPayment",
            "timelock": 0,
            "timeout": 0,
            "contractName": "EscrowPayment",
            "functionName": "fulfill",
            "parameters": [
                {
                    "name": "_did",
                    "type": "bytes32",
                    "value": "{parameter.assetId}"
                },
                {
                    "name": "_amounts",
                    "type": "uint256[]",
                    "value": []
                },
                {
                    "name": "_receivers",
                    "type": "address[]",
                    "value": []
                },
                {
                    "name": "_sender",
                    "type": "address",
                    "value": ""
                },
                {
                    "name": "_tokenAddress",
                    "type": "address",
                    "value": ""
                },
                {
                    "name": "_lockCondition",
                    "type": "bytes32",
                    "value": "{contract.LockPaymentCondition.address}"
                },
                {
                    "name": "_releaseCondition",
                    "type": "bytes32",
                    "value": "{contract.ExecComputeCondition.address}"
                }
            ],
            "events": [
                {
                    "name": "Fulfilled",
                    "actorType": "publisher",
                    "handler": {
                        "moduleName": "escrowPaymentCondition",
                        "functionName": "verifyRewardTokens",
                        "version": "0.1"
                    }
                }
            ]
        }
    ]
  }
}


import logging
import secrets

import pytest
from contracts_lib_py import Keeper
from contracts_lib_py.exceptions import (
    DIDNotFound,
)
from web3 import Web3

from common_utils_py.did import DID, did_to_id
from common_utils_py.did_resolver.did_resolver import (
    DIDResolver,
)
from common_utils_py.utils.utilities import generate_prefixed_id

logger = logging.getLogger()


def keeper():
    return Keeper.get_instance()


def test_did_resolver_library(publisher_account, metadata_instance, ddo_sample_2):
    did_registry = keeper().did_registry
    checksum_test = Web3.keccak(text='checksum')
    value_test = metadata_instance.root_url

    did_resolver = DIDResolver(keeper().did_registry)

    did_seed = generate_prefixed_id()
    asset_id = did_registry.hash_did(did_seed, publisher_account.address)

    asset1 = ddo_sample_2
    asset1._did = DID.did(asset_id)
    did_registry.register(did_seed, checksum_test, url=value_test, account=publisher_account)
    metadata_instance.publish_asset_ddo(asset1)

    did_resolved = did_resolver.resolve(asset1.did)
    assert did_resolved
    assert did_resolved.did == asset1.did

    with pytest.raises(ValueError):
        did_resolver.resolve(asset1.asset_id)

    metadata_instance.retire_asset_ddo(asset1.did)


def test_did_not_found():
    did_resolver = DIDResolver(keeper().did_registry)
    did_id = secrets.token_hex(32)
    did_id_bytes = Web3.toBytes(hexstr=did_id)
    with pytest.raises(DIDNotFound):
        did_resolver.resolve(did_id_bytes)


def test_get_resolve_url(metadata_instance, publisher_account):
    register_account = publisher_account
    did_registry = keeper().did_registry

    did_seed = generate_prefixed_id()
    asset_id = did_registry.hash_did(did_seed, register_account.address)

    did = DID.did(asset_id)

    value_test = metadata_instance.root_url
    did_resolver = DIDResolver(keeper().did_registry)
    did_registry.register(did_seed, b'test', url=value_test, account=register_account)
    url = did_resolver.get_resolve_url(Web3.toBytes(hexstr=asset_id))
    assert url == value_test


def test_get_resolve_multiple_urls(publisher_account):
    register_account = publisher_account
    did_registry = keeper().did_registry
    did_resolver = DIDResolver(keeper().did_registry)

    iterations = 3
    counter = 0

    while counter < iterations:
        _seed = generate_prefixed_id()
        _asset_id = did_registry.hash_did(_seed, register_account.address)
        _url = 'http://localhost:5000/' + str(counter)
        did_registry.register(_seed, b'test', url=_url, account=register_account)
        assert did_resolver.get_resolve_url(Web3.toBytes(hexstr=_asset_id)) == _url
        counter = counter + 1


def test_get_did_not_valid():
    did_resolver = DIDResolver(keeper().did_registry)
    with pytest.raises(TypeError):
        did_resolver.get_resolve_url('not valid')



from common_utils_py.agreements.service_agreement import ServiceAgreement
from common_utils_py.ddo.service import Service


def test_store_and_retrieve_service(metadata_instance, json_service):
    agreement_id = ServiceAgreement.create_new_agreement_id()
    service = Service.from_json(json_service)

    # store the service agreement
    result = metadata_instance.store_service_agreement(agreement_id, service)
    assert result is True

    result = metadata_instance.get_service_agreement(agreement_id)
    assert result.type == service.type
    assert result.index == service.index
    assert result.service_endpoint == service.service_endpoint
    assert result.attributes == service.attributes



def test_encoding(setup_basic_environment):
    """Test the agreement flow according to the AccessTemplate."""
    (
        keeper,
        publisher_acc,
        consumer_acc,
    ) = setup_basic_environment

    agreement_id = '0x6098b775850142a1bd13ceca600a1dfad5c58aae6cd8499c9f02c0cf1e052a17'
    token_address = '0x80163Dec819063F87ad4F1b6A24DA551C93c5777'
    amounts = [10, 2]
    receivers = ['0x00Bd138aBD70e2F00903268F3Db08f2D25677C9e', '0x068Ed00cF0441e4829D9784fCBe7b9e26D4BD8d0']
    asset_id = '0x0792cbe4ea0285a2a2db26f262c44b7478e02dc3da380f7c54be1634300b86b7'
    escrow_condition_address = '0x602e6AA9Ea3293410B6103283cE109E2073FC160'

    _contract_lock_hash = '0x' + keeper.lock_payment_condition.hash_values(asset_id, escrow_condition_address, token_address, amounts, receivers).hex()
    assert _contract_lock_hash == '0x20d5d330cad3f7d6769da69ac3f569676c83db2c27e750e0c65c43be359b059c'

    _contract_lock_hash_2 = '0x' + keeper.lock_payment_condition.contract.functions.hashValues(asset_id, escrow_condition_address, token_address, amounts, receivers).call().hex()
    assert _contract_lock_hash == _contract_lock_hash_2

    _id = '0x' + keeper.lock_payment_condition.contract.functions.generateId(agreement_id, _contract_lock_hash).call().hex()

    assert _id != _contract_lock_hash



from ecies import encrypt, decrypt
from ecies.utils import generate_eth_key

from common_utils_py.utils.crypto import decryption, encryption, get_keys_from_file, get_rsa_public_key_from_file, \
    get_rsa_private_key_from_file, rsa_encryption, rsa_decryption, aes_encryption, aes_decryption

PROVIDER_KEY_FILE = 'tests/resources/data/publisher_key_file.json'
PROVIDER_PASSWORD = 'node0'
PROVIDER_ADDRESS = '0x00bd138abd70e2f00903268f3db08f2d25677c9e'
RSA_PRIVKEY_FILE = 'tests/resources/data/rsa_priv_key.pem'
RSA_PUBKEY_FILE = 'tests/resources/data/rsa_pub_key.pem'


def test_encryption_decryption():
    eth_k = generate_eth_key()
    private_key_hex = eth_k.to_hex()  # hex string
    public_key_hex = eth_k.public_key.to_hex()  # hex string
    data = 'hi there'
    result = decrypt(private_key_hex, encrypt(public_key_hex, data.encode()))
    print(result)
    print('decoded=' + result.decode())
    assert result == data.encode()
    assert data == result.decode()


def test_encryption_decryption_with_credentials():
    keyfile_path = PROVIDER_KEY_FILE
    keyfile_password = PROVIDER_PASSWORD

    print('KeyFile Path = ' + keyfile_path)

    (public_key_hex, private_key_hex) = get_keys_from_file(keyfile_path, keyfile_password)

    data = b'hi there'
    assert data == decryption(private_key_hex, encryption(public_key_hex, data))
    assert b'it should fail' != decryption(private_key_hex, encryption(public_key_hex, b'kdas'))


def test_rsa_encryption_decryption():
    pub_key = get_rsa_public_key_from_file(RSA_PUBKEY_FILE)
    priv_key = get_rsa_private_key_from_file(RSA_PRIVKEY_FILE)

    data = b'hi there'
    encrypted_data, encrypted_aes_key = rsa_encryption(pub_key, data)
    assert data == rsa_decryption(priv_key, encrypted_data, encrypted_aes_key)

    encrypted_data, encrypted_aes_key = rsa_encryption(pub_key, b'kdas')
    assert b'it should fail' != rsa_decryption(priv_key, encrypted_data, encrypted_aes_key)


def test_rsa_encryption_decryption_long():
    pub_key = get_rsa_public_key_from_file(RSA_PUBKEY_FILE)
    priv_key = get_rsa_private_key_from_file(RSA_PRIVKEY_FILE)

    data = b'hi there, this is a much longer message that should failing during the encryption if you' \
           b'dont use an alternative approach. So this test tries to check the new flow works with ' \
           b'longer messages'
    encrypted_data, encrypted_aes_key = rsa_encryption(pub_key, data)
    assert data == rsa_decryption(priv_key, encrypted_data, encrypted_aes_key)


def test_aes_encryption_decryption():
    passphrase = 'my passphrase'

    data = b'hi there, this is a much longer message that should failing during the encryption if you' \
           b'dont use an alternative approach. So this test tries to check the new flow works with ' \
           b'longer messages'
    encrypted_data = aes_encryption(data, passphrase)
    assert data == aes_decryption(encrypted_data, passphrase)





import logging

from common_utils_py.metadata.metadata_provider import MetadataProvider
from common_utils_py.did import did_to_id_bytes

logger = logging.getLogger('keeper')


class DIDResolver:
    """
    DID Resolver class
    Resolve DID to a URL/DDO.
    """

    def __init__(self, did_registry):
        self._did_registry = did_registry

    def resolve(self, did):
        """
        Resolve a DID to an URL/DDO or later an internal/external DID.

        :param did: 32 byte value or DID string to resolver, this is part of the ocean
            DID did:nv:<32 byte value>
        :return string: URL or DDO of the resolved DID
        :return None: if the DID cannot be resolved
        :raises ValueError: if did is invalid
        :raises TypeError: if did has invalid format
        :raises TypeError: on non 32byte value as the DID
        :raises TypeError: on any of the resolved values are not string/DID bytes.
        :raises OceanDIDNotFound: if no DID can be found to resolve.
        """

        did_bytes = did_to_id_bytes(did)
        if not isinstance(did_bytes, bytes):
            raise TypeError('Invalid did: a 32 Byte DID value required.')

        # resolve a DID to a DDO
        url = self.get_resolve_url(did_bytes)
        logger.debug(f'found did {did} -> url={url}')
        return MetadataProvider.get_metadata_provider(url).get_asset_ddo(did)

    def get_resolve_url(self, did_bytes):
        """Return a did value and value type from the block chain event record using 'did'.

        :param did_bytes: DID, hex-str
        :return url: Url, str
        """
        data = self._did_registry.get_registered_attribute(did_bytes)
        if not (data and data.get('value')):
            return None

        return data['value']



from common_utils_py.agreements.service_agreement_condition import Event, ServiceAgreementCondition


class ServiceAgreementTemplate(object):
    """Class representing a Service Agreement Template."""
    TEMPLATE_ID_KEY = 'templateId'

    def __init__(self, template_id=None, name=None, creator=None, template_json=None):
        self.template_id = template_id
        self.name = name
        self.creator = creator
        self.template = {}
        if template_json:
            self.parse_template_json(template_json)

    def parse_template_json(self, template_json):
        """
        Parse a template from a json.

        :param template_json: json dict
        """
        if 'template' in template_json['serviceAgreementTemplate']:
            self.template = template_json['serviceAgreementTemplate']['template']
        else:
            self.template = template_json['serviceAgreementTemplate']

    def set_template_id(self, template_id):
        """
        Assign the template id to the template.

        :param template_id: string
        """
        self.template_id = template_id

    @property
    def fulfillment_order(self):
        """
        List with the fulfillment order.

        :return: list
        """
        return self.template['fulfillmentOrder']

    @property
    def condition_dependency(self):
        """
        Dictionary with the dependencies of the conditions.

        :return: dict
        """
        return self.template['conditionDependency']

    @property
    def contract_name(self):
        """
        Contract name of the template.

        :return: string
        """
        return self.template['contractName']

    @property
    def agreement_events(self):
        """
        List of agreements events.

        :return: list
        """
        return [Event(e) for e in self.template['events']]

    @property
    def conditions(self):
        """
        List of conditions.

        :return: list
        """
        return [
            ServiceAgreementCondition(cond_json) for cond_json in self.template['conditions']
        ]

    def set_conditions(self, conditions):
        """
        Set the conditions of the template.

        :param conditions: list of conditions.
        """
        self.template['conditions'] = [cond.as_dictionary() for cond in conditions]

    def get_event_to_args_map(self, contract_by_name):
        """
        keys in returned dict have the format <contract_name>.<event_name>
        """
        cond_contract_tuples = [(cond, contract_by_name[cond.contract_name]) for cond in
                                self.conditions]
        event_to_args = {
            f'{cond.contract_name}.{cond.events[0].name}': (
                contract.get_event_argument_names(cond.events[0].name)
            )
            for cond, contract in cond_contract_tuples
        }
        agr_event_key = f'{self.contract_name}.{self.agreement_events[0].name}'
        event_to_args[agr_event_key] = contract_by_name[
            self.contract_name].get_event_argument_names(
            self.agreement_events[0].name
        )

        return event_to_args

    def as_dictionary(self):
        """
        Return the service agreement template as a dictionary.

        :return: dict
        """
        template = {
            'contractName': self.contract_name,
            'events': [e.as_dictionary() for e in self.agreement_events],
            'fulfillmentOrder': self.fulfillment_order,
            'conditionDependency': self.condition_dependency,
            'conditions': [cond.as_dictionary() for cond in self.conditions]
        }
        return {
            # 'type': self.DOCUMENT_TYPE,
            'name': self.name,
            'creator': self.creator,
            'serviceAgreementTemplate': template
        }



"""
    Test did_lib
"""

import pytest
from contracts_lib_py import Keeper

from common_utils_py.agreements.service_agreement import ServiceTypes
from common_utils_py.ddo.ddo import DDO
from common_utils_py.ddo.public_key_base import PublicKeyBase
from common_utils_py.ddo.public_key_rsa import PUBLIC_KEY_TYPE_ETHEREUM_ECDSA, PUBLIC_KEY_TYPE_RSA
from common_utils_py.did import DID, did_to_id_bytes
from common_utils_py.utils.utilities import checksum
from tests.resources.helper_functions import get_publisher_account

TEST_SERVICE_TYPE = 'ocean-meta-storage'
TEST_SERVICE_URL = 'http://localhost:8005'


def test_create_ddo(metadata):
    pub_acc = get_publisher_account()
    ddo = DDO()
    ddo.add_service(ServiceTypes.METADATA, 'http://mymetadata.com',
                    values=dict({"attributes": metadata}), index='0')
    checksums = dict()
    for service in ddo.services:
        checksums[str(service.index)] = checksum(service.main)
    ddo.add_proof(checksums, pub_acc)
    did = ddo.assign_did(DID.encoded_did(ddo.proof['checksum']))
    ddo.proof['signatureValue'] = Keeper.sign_hash(did_to_id_bytes(did), pub_acc)
    ddo.add_public_key(did, pub_acc.address)
    ddo.add_authentication(did, PUBLIC_KEY_TYPE_RSA)


def test_creating_ddo_from_scratch():
    # create an empty ddo
    ddo = DDO()
    assert ddo.did is None
    assert ddo.asset_id is None
    assert ddo.created is not None

    did = DID.did("0x99999999999999999")
    ddo.assign_did(did)
    assert ddo.did == did

    pub_acc = get_publisher_account()

    ddo.add_service(TEST_SERVICE_TYPE, TEST_SERVICE_URL)

    # add a proof to the first public_key/authentication
    ddo.add_proof('checksum', pub_acc)
    ddo_text_proof = ddo.as_text()
    assert ddo_text_proof

    pub_acc = get_publisher_account()
    assert not ddo.public_keys
    ddo.add_public_key(did, pub_acc.address)
    assert len(ddo.public_keys) == 1
    assert ddo.get_public_key(0) == ddo.public_keys[0]
    with pytest.raises(IndexError):
        ddo.get_public_key(1)

    assert ddo.get_public_key(did) == ddo.public_keys[0]
    assert ddo.get_public_key('0x32233') is None

    assert not ddo.authentications
    ddo.add_authentication(did, '')
    assert len(ddo.authentications) == 1


def test_create_auth_from_json():
    auth = {'publicKey': '0x00000', 'type': 'auth-type', 'nothing': ''}
    assert (
            DDO.create_authentication_from_json(auth)
            ==
            {'publicKey': '0x00000', 'type': 'auth-type'}
    )
    with pytest.raises(ValueError):
        DDO.create_authentication_from_json({'type': 'auth-type'})


def test_create_public_key_from_json():
    pkey = {'id': 'pkeyid', 'type': 'keytype', 'owner': '0x00009'}
    pub_key_inst = DDO.create_public_key_from_json(pkey)
    assert isinstance(pub_key_inst, PublicKeyBase)
    assert pub_key_inst.get_id() == pkey['id']
    assert pub_key_inst.get_type() == PUBLIC_KEY_TYPE_ETHEREUM_ECDSA
    assert pub_key_inst.get_owner() == pkey['owner']

    pub_key_inst = DDO.create_public_key_from_json(
        {'type': PUBLIC_KEY_TYPE_RSA}
    )
    assert pub_key_inst.get_id() == ''
    assert pub_key_inst.get_type() == PUBLIC_KEY_TYPE_RSA
    assert pub_key_inst.get_owner() is None


def test_load_ddo_json(ddo_sample_2):
    service = ddo_sample_2.get_service(ServiceTypes.METADATA)
    assert service
    assert service.type == ServiceTypes.METADATA
    assert service.attributes


def test_ddo_dict(ddo_sample_2):
    assert len(ddo_sample_2.public_keys) == 1
    assert ddo_sample_2.did == \
           'did:nv:0c184915b07b44c888d468be85a9b28253e80070e5294b1aaed81c2f0264e431'


def test_find_service(ddo_sample):
    service = ddo_sample.get_service_by_index(0)
    assert service and service.type == ServiceTypes.METADATA, 'Failed to find service by integer ' \
                                                              'id.'
    service = ddo_sample.get_service_by_index('0')
    assert service and service.type == ServiceTypes.METADATA, 'Failed to find service by str(int)' \
                                                              ' id.'

    service = ddo_sample.get_service_by_index(3)
    assert service and service.type == ServiceTypes.ASSET_ACCESS, 'Failed to find service by ' \
                                                                  'integer id.'
    service = ddo_sample.get_service_by_index('3')
    assert service and service.type == ServiceTypes.ASSET_ACCESS, 'Failed to find service by str(' \
                                                                  'int) id.'

    service = ddo_sample.get_service(ServiceTypes.ASSET_ACCESS)
    assert service and service.type == ServiceTypes.ASSET_ACCESS, 'Failed to find service by id ' \
                                                                  'using service type.'
    assert service.index == 3, 'index not as expected.'

    service = ddo_sample.get_service(ServiceTypes.METADATA)
    assert service and service.type == ServiceTypes.METADATA, 'Failed to find service by id using ' \
                                                              '' \
                                                              '' \
                                                              'service ' \
                                                              'type.'
    assert service.index == 0, 'index not as expected.'



"""DID Lib to do DID's and DDO's."""
import re

from eth_utils import add_0x_prefix, remove_0x_prefix
from web3 import Web3

from common_utils_py.utils.utilities import checksum

NEVERMINED_PREFIX = 'did:nv:'


class DID:
    """Class representing an asset DID."""

    @staticmethod
    def did(asset_id):
        """
        Create a did.

        Format of the did:
        did:nv:cb36cf78d87f4ce4a784f17c2a4a694f19f3fbf05b814ac6b0b7197163888865

        :param asset_id: the asset id
        :return: Asset did, str.
        """
        return NEVERMINED_PREFIX + remove_0x_prefix(asset_id).zfill(64)

    @staticmethod
    def encoded_did(seed):
        """
        Create a did.

        Format of the did:
        did:nv:cb36cf78d87f4ce4a784f17c2a4a694f19f3fbf05b814ac6b0b7197163888865

        :param seed: The list of checksums that is allocated in the proof, dict
        :return: Asset did, str.
        """
        return NEVERMINED_PREFIX + remove_0x_prefix(checksum(seed))

def did_parse(did):
    """
    Parse a DID into it's parts.

    :param did: Asset did, str.
    :return: Python dictionary with the method and the id.
    """
    if not isinstance(did, str):
        raise TypeError(f'Expecting DID of string type, got {did} of {type(did)} type')

    match = re.match('^did:([a-z0-9]+):([a-zA-Z0-9-.]+)(.*)', did)
    if not match:
        raise ValueError(f'DID {did} does not seem to be valid.')

    result = {
        'method': match.group(1),
        'id': match.group(2),
    }

    return result


def is_did_valid(did):
    """
    Did validator.

    Return True if the did is a valid DID with the method name 'op' and the id
    in the Nevermined format

    :param did: Asset did, str
    :return bool
    """
    result = did_parse(did)
    if result:
        return result['id'] is not None
    return False


def id_to_did(did_id, method='nv'):
    """Return an Nevermined DID from given a hex id."""
    if isinstance(did_id, bytes):
        did_id = Web3.toHex(did_id)

    # remove leading '0x' of a hex string
    if isinstance(did_id, str):
        did_id = remove_0x_prefix(did_id)
    else:
        raise TypeError("did id must be a hex string or bytes")

    # test for zero address
    if Web3.toBytes(hexstr=did_id) == b'':
        did_id = '0'
    return f'did:{method}:{did_id}'


def did_to_id(did):
    """Return an id extracted from a DID string."""
    result = did_parse(did)
    if result and result['id'] is not None:
        return result['id']
    return None


def did_to_id_bytes(did):
    """
    Return an DID to it's corresponding hex id in bytes.

    So did:nv:<hex>, will return <hex> in byte format
    """
    if isinstance(did, str):
        if re.match('^[0x]?[0-9A-Za-z]+$', did):
            raise ValueError(f'{did} must be a DID not a hex string')
        else:
            did_result = did_parse(did)
            if not did_result:
                raise ValueError(f'{did} is not a valid did')
            if not did_result['id']:
                raise ValueError(f'{did} is not a valid ocean did')
            id_bytes = Web3.toBytes(hexstr=did_result['id'])
    elif isinstance(did, bytes):
        id_bytes = did
    else:
        raise TypeError(
            f'Unknown did format, expected str or bytes, got {did} of type {type(did)}'
        )
    return id_bytes


def convert_to_bytes(did):
    if isinstance(did, str):
        if re.match('^[0x]?[0-9A-Za-z]+$', did):
            return Web3.toBytes(hexstr=add_0x_prefix(did))
        else:
            return Web3.toBytes(hexstr=add_0x_prefix(did_to_id(did)))
    elif isinstance(did, bytes):
        return did
    else:
        raise ValueError(f'The id {did} is not in a valid format.')



"""Metadata module."""
from .metadata import Metadata


class MetadataProvider:
    """Provides the Metadata instance."""
    _metadata_class = Metadata

    @staticmethod
    def get_metadata_provider(url):
        """ Get an Metadata instance."""
        return MetadataProvider._metadata_class(url)

    @staticmethod
    def set_metadata_class(metadata_class):
        """
         Set an Metadata class

        :param metadata_class: Metadata or similar compatible class
        """
        MetadataProvider._metadata_class = metadata_class



from tests.resources.helper_functions import log_event


def test_nft_sales_flow(setup_nft_sales_agreements_environment):
    """Test the DID Sales agreement flow according to the DIDSalesTemplate."""
    (
        keeper,
        ddo,
        publisher_acc,
        consumer_acc,
        agreement_id,
        nft_access_agreement_id,
        asset_id,
        price,
        service_agreement,
        nft_access_service_agreement,
        (lock_cond_id, access_cond_id, escrow_cond_id),
        (nft_access_cond_id, nft_holder_cond_id)
    ) = setup_nft_sales_agreements_environment

    print('creating NFT Sales Agreement:'
          'agrId: ', agreement_id,
          'asset_id', asset_id,
          '[lock_cond_id, access_cond_id, escrow_cond_id]',
          [lock_cond_id, access_cond_id, escrow_cond_id],
          'tlocks', service_agreement.conditions_timelocks,
          'touts', service_agreement.conditions_timeouts,
          'consumer', consumer_acc.address,
          'publisher', publisher_acc.address
          )

    amounts = service_agreement.get_amounts_int()
    receivers = service_agreement.get_receivers()
    number_nfts = service_agreement.get_number_nfts()
    token_address = keeper.token.address

    receiver_0_starting_balance = keeper.token.get_token_balance(
        keeper.agreement_manager.to_checksum_address(receivers[0]))

    assert keeper.template_manager.is_template_approved(
        keeper.nft_sales_template.address), 'Template is not approved.'
    assert keeper.did_registry.get_block_number_updated(asset_id) > 0, 'asset id not registered'
    success = keeper.nft_sales_template.create_agreement(
        agreement_id,
        asset_id,
        [lock_cond_id, access_cond_id, escrow_cond_id],
        service_agreement.conditions_timelocks,
        service_agreement.conditions_timeouts,
        consumer_acc.address,
        publisher_acc
    )
    print('create agreement: ', success)
    assert success, f'createAgreement failed {success}'
    event = keeper.nft_sales_template.subscribe_agreement_created(
        agreement_id,
        10,
        log_event(keeper.nft_sales_template.AGREEMENT_CREATED_EVENT),
        (),
        wait=True
    )
    assert event, 'no event for AgreementCreated '

    # Verify condition types (condition contracts)
    agreement = keeper.agreement_manager.get_agreement(agreement_id)
    assert agreement.did == asset_id, ''
    cond_types = keeper.nft_sales_template.get_condition_types()
    for i, cond_id in enumerate(agreement.condition_ids):
        cond = keeper.condition_manager.get_condition(cond_id)
        assert cond.type_ref == cond_types[i]
        assert int(cond.state) == 1

    # Give consumer some tokens
    keeper.dispenser.request_vodkas(price, consumer_acc)

    # Fulfill lock_payment_condition
    pub_token_balance = keeper.token.get_token_balance(publisher_acc.address)
    consumer_token_balance = keeper.token.get_token_balance(consumer_acc.address)
    assert consumer_token_balance >= price, 'Consumer doesnt have enough tokens to afford the purchase'

    starting_balance = keeper.token.get_token_balance(keeper.escrow_payment_condition.address)
    keeper.token.token_approve(keeper.lock_payment_condition.address, price, consumer_acc)
    keeper.token.token_approve(keeper.escrow_payment_condition.address, price, consumer_acc)

    tx_hash = keeper.lock_payment_condition.fulfill(
        agreement_id,
        asset_id,
        keeper.escrow_payment_condition.address,
        token_address,
        amounts,
        receivers,
        consumer_acc)
    keeper.lock_payment_condition.get_tx_receipt(tx_hash)
    event = keeper.lock_payment_condition.subscribe_condition_fulfilled(
        agreement_id,
        10,
        log_event(keeper.lock_payment_condition.FULFILLED_EVENT),
        (),
        wait=True
    )
    assert event, 'no event for LockPaymentCondition.Fulfilled'
    assert keeper.condition_manager.get_condition_state(lock_cond_id) == 2, ''
    assert keeper.token.get_token_balance(
        keeper.escrow_payment_condition.address
    ) == (price + starting_balance), ''

    assert keeper.nft_upgradeable.balance(publisher_acc.address, asset_id) >= number_nfts

    assert keeper.nft_upgradeable.is_approved_for_all(publisher_acc.address,
                                                       keeper.transfer_nft_condition.address) is True

    # Fulfill transfer did condition
    tx_hash = keeper.transfer_nft_condition.fulfill(
        agreement_id,
        asset_id,
        consumer_acc.address,
        number_nfts,
        lock_cond_id,
        publisher_acc
    )
    keeper.transfer_nft_condition.get_tx_receipt(tx_hash)
    event = keeper.transfer_nft_condition.subscribe_condition_fulfilled(
        agreement_id,
        20,
        log_event(keeper.transfer_nft_condition.FULFILLED_EVENT),
        (),
        wait=True
    )
    assert event, 'no event for TransferNFTCondition.Fulfilled'
    assert keeper.condition_manager.get_condition_state(access_cond_id) == 2, ''

    # Fulfill escrow_payment_condition
    tx_hash = keeper.escrow_payment_condition.fulfill(
        agreement_id, asset_id, amounts, receivers,
        keeper.escrow_payment_condition.address, token_address, lock_cond_id,
        access_cond_id, publisher_acc
    )

    keeper.escrow_payment_condition.get_tx_receipt(tx_hash)
    event = keeper.escrow_payment_condition.subscribe_condition_fulfilled(
        agreement_id,
        10,
        log_event(keeper.escrow_payment_condition.FULFILLED_EVENT),
        (),
        wait=True
    )

    assert event, 'no event for EscrowPayment.Fulfilled'
    assert keeper.condition_manager.get_condition_state(escrow_cond_id) == 2, ''
    assert keeper.token.get_token_balance(
        keeper.escrow_payment_condition.address
    ) == starting_balance, ''

    assert keeper.token.get_token_balance(
        keeper.agreement_manager.to_checksum_address(receivers[0])) == (
                       receiver_0_starting_balance + int(amounts[0])), ''


    ### NFT ACCESS FLOW
    print('Downloading NFT content, DID: ' + ddo.did)

    success = keeper.nft_access_template.create_agreement(
        nft_access_agreement_id,
        asset_id,
        [nft_holder_cond_id, nft_access_cond_id],
        nft_access_service_agreement.conditions_timelocks,
        nft_access_service_agreement.conditions_timeouts,
        consumer_acc.address,
        publisher_acc
    )
    print('Create NFT Access agreement: ', success)
    assert success, f'createAgreement failed {success}'
    event = keeper.nft_access_template.subscribe_agreement_created(
        nft_access_agreement_id,
        10,
        log_event(keeper.nft_access_template.AGREEMENT_CREATED_EVENT),
        (),
        wait=True
    )
    assert event, 'no event for AgreementCreated '

    # Fulfill nft_holder_condition
    tx_hash = keeper.nft_holder_condition.fulfill(
        nft_access_agreement_id, asset_id, consumer_acc.address, number_nfts, publisher_acc
    )

    keeper.nft_holder_condition.get_tx_receipt(tx_hash)
    event = keeper.nft_holder_condition.subscribe_condition_fulfilled(
        nft_access_agreement_id,
        10,
        log_event(keeper.nft_holder_condition.FULFILLED_EVENT),
        (),
        wait=True
    )

    tx_hash = keeper.nft_access_condition.fulfill(
        nft_access_agreement_id, asset_id, consumer_acc.address, publisher_acc
    )
    keeper.nft_access_condition.get_tx_receipt(tx_hash)
    event = keeper.nft_access_condition.subscribe_condition_fulfilled(
        nft_access_agreement_id,
        20,
        log_event(keeper.nft_access_condition.FULFILLED_EVENT),
        (),
        wait=True
    )
    assert event, 'no event for AccessCondition.Fulfilled'
    assert keeper.condition_manager.get_condition_state(access_cond_id) == 2, ''



import pytest

from common_utils_py.utils import keytransfer


def test_poseidon():
    assert keytransfer.poseidon([1,2]) == 7853200120776062878684798364095072458815029376092732009249414926327459813530

def test_mimc():
    assert keytransfer.mimc(1, 2, 8496618697356220059886051648941066104102428018438044414794308085967084497473) == [17405299444732615738647569185121043774784715854244839029575270519342275455504, 12697608136635457114705488273155937541518518394149826673715293889909126944425]

def test_jubjub():
    buyer_k = 123
    provider_k = 234
    buyer_pub = keytransfer.mulPointEscalar(keytransfer.base8, buyer_k)
    provider_pub = keytransfer.mulPointEscalar(keytransfer.base8, provider_k)
    assert keytransfer.mulPointEscalar(buyer_pub, provider_k) == keytransfer.mulPointEscalar(provider_pub, buyer_k)

def test_prover():
    data = b"123456789q01234567890q1234567890"
    buyer_k = 123
    provider_k = 234
    buyer_pub = keytransfer.mulPointEscalar(keytransfer.base8, buyer_k)

    prover = keytransfer.make_prover("/usr/local/share/keytransfer/keytransfer.zkey", "/usr/local/share/keytransfer/keytransfer.dat")
    res = keytransfer.prove_transfer(prover, buyer_pub, provider_k, data)

def test_signature():
    data = b"123456789q"
    buyer_secret = "abc123"
    buyer_k = keytransfer.make_key(buyer_secret)
    dta_num = int(data.hex(), 16)
    buyer_pub = keytransfer.mulPointEscalar(keytransfer.base8, buyer_k)
    sig = keytransfer.sign(buyer_secret, dta_num)
    assert keytransfer.verify(buyer_pub, dta_num, sig)




from contracts_lib_py.utils import generate_multi_value_hash
from web3 import Web3
from web3._utils.personal import new_account

from common_utils_py.agreements.service_agreement import ServiceAgreement, ServiceTypes
from tests.resources.helper_functions import (
    log_event
)


def test_access_template_flow(setup_agreements_environment):
    """Test the agreement flow according to the AccessTemplate."""
    (
        keeper,
        ddo,
        publisher_acc,
        consumer_acc,
        agreement_id,
        asset_id,
        price,
        service_agreement,
        (lock_cond_id, access_cond_id, escrow_cond_id),

    ) = setup_agreements_environment

    print('creating agreement:'
          'agrId: ', agreement_id,
          'asset_id', asset_id,
          '[lock_cond_id, access_cond_id, escrow_cond_id]',
          [lock_cond_id, access_cond_id, escrow_cond_id],
          'tlocks', service_agreement.conditions_timelocks,
          'touts', service_agreement.conditions_timeouts,
          'consumer', consumer_acc.address,
          'publisher', publisher_acc.address
          )

    amounts = service_agreement.get_amounts_int()
    receivers = service_agreement.get_receivers()
    # receivers = [publisher_acc.address, '0x2B5AD5c4795c026514f8317c7a215E218DcCD6cF']
    token_address = keeper.token.address

    receiver_0_starting_balance = keeper.token.get_token_balance(
        keeper.agreement_manager.to_checksum_address(receivers[0]))

    assert keeper.template_manager.is_template_approved(
        keeper.access_template.address), 'Template is not approved.'
    assert keeper.did_registry.get_block_number_updated(asset_id) > 0, 'asset id not registered'
    success = keeper.access_template.create_agreement(
        agreement_id,
        asset_id,
        [access_cond_id, lock_cond_id, escrow_cond_id],
        service_agreement.conditions_timelocks,
        service_agreement.conditions_timeouts,
        consumer_acc.address,
        publisher_acc
    )
    print('create agreement: ', success)
    assert success, f'createAgreement failed {success}'
    event = keeper.access_template.subscribe_agreement_created(
        agreement_id,
        10,
        log_event(keeper.access_template.AGREEMENT_CREATED_EVENT),
        (),
        wait=True
    )
    assert event, 'no event for AgreementCreated '

    # Verify condition types (condition contracts)
    agreement = keeper.agreement_manager.get_agreement(agreement_id)
    assert agreement.did == asset_id, ''
    cond_types = keeper.access_template.get_condition_types()
    for i, cond_id in enumerate(agreement.condition_ids):
        cond = keeper.condition_manager.get_condition(cond_id)
        assert cond.type_ref == cond_types[i]
        assert int(cond.state) == 1

    # Give consumer some tokens
    keeper.dispenser.request_vodkas(price, consumer_acc)

    # Fulfill lock_payment_condition
    pub_token_balance = keeper.token.get_token_balance(publisher_acc.address)
    starting_balance = keeper.token.get_token_balance(keeper.escrow_payment_condition.address)
    keeper.token.token_approve(keeper.lock_payment_condition.address, price, consumer_acc)
    tx_hash = keeper.lock_payment_condition.fulfill(
        agreement_id, asset_id, keeper.escrow_payment_condition.address, token_address, amounts, receivers, consumer_acc)
    keeper.lock_payment_condition.get_tx_receipt(tx_hash)
    event = keeper.lock_payment_condition.subscribe_condition_fulfilled(
        agreement_id,
        10,
        log_event(keeper.lock_payment_condition.FULFILLED_EVENT),
        (),
        wait=True
    )
    assert event, 'no event for LockPaymentCondition.Fulfilled'
    assert keeper.condition_manager.get_condition_state(lock_cond_id) == 2, ''
    assert keeper.token.get_token_balance(
        keeper.escrow_payment_condition.address
    ) == (price + starting_balance), ''

    # Fulfill access_condition
    tx_hash = keeper.access_condition.fulfill(
        agreement_id, asset_id, consumer_acc.address, publisher_acc
    )
    keeper.access_condition.get_tx_receipt(tx_hash)
    event = keeper.access_condition.subscribe_condition_fulfilled(
        agreement_id,
        20,
        log_event(keeper.access_condition.FULFILLED_EVENT),
        (),
        wait=True
    )
    assert event, 'no event for AccessCondition.Fulfilled'
    assert keeper.condition_manager.get_condition_state(access_cond_id) == 2, ''

    # Fulfill escrow_payment_condition
    tx_hash = keeper.escrow_payment_condition.fulfill(
        agreement_id, asset_id, amounts, receivers,
        keeper.escrow_payment_condition.address, token_address, lock_cond_id,
        access_cond_id, publisher_acc
    )

    keeper.escrow_payment_condition.get_tx_receipt(tx_hash)
    event = keeper.escrow_payment_condition.subscribe_condition_fulfilled(
        agreement_id,
        10,
        log_event(keeper.escrow_payment_condition.FULFILLED_EVENT),
        (),
        wait=True
    )

    assert event, 'no event for EscrowPayment.Fulfilled'
    assert keeper.condition_manager.get_condition_state(escrow_cond_id) == 2, ''
    assert keeper.token.get_token_balance(
        keeper.escrow_payment_condition.address
    ) == starting_balance, ''

    assert keeper.token.get_token_balance(
        keeper.agreement_manager.to_checksum_address(receivers[0])) == (receiver_0_starting_balance + int(amounts[0])), ''


def test_agreement_hash(ddo_sample):
    """
    This test verifies generating agreement hash using fixed inputs and ddo example.
    This will make it easier to compare the hash generated from different languages.
    """
    w3 = Web3
    did = "did:nv:0c184915b07b44c888d468be85a9b28253e80070e5294b1aaed81c2f0264e430"
    template_id = w3.toChecksumAddress("0x00bd138abd70e2f00903268f3db08f2d25677c9e")
    agreement_id = '0xf136d6fadecb48fdb2fc1fb420f5a5d1c32d22d9424e47ab9461556e058fefaa'
    sa = ServiceAgreement.from_service_dict(
        ddo_sample.get_service(ServiceTypes.ASSET_ACCESS).as_dictionary())
    sa.service_agreement_template.set_template_id(template_id)
    assert template_id == sa.template_id, ''
    assert did == ddo_sample.did
    # Don't generate condition ids, use fixed ids so we get consistent hash
    # (access_id, lock_id, escrow_id) = sa.generate_agreement_condition_ids(
    #     agreement_id, ddo.asset_id, consumer, publisher, keeper)
    access_id = '0x2d7c1d60dc0c3f52aa9bd71ffdbe434a0e58435571e64c893bc9646fea7f6ec1'
    lock_id = '0x1e265c434c14e668695dda1555088f0ea4356f596bdecb8058812e7dcba9ee73'
    escrow_id = '0xe487fa6d435c2f09ef14b65b34521302f1532ac82ba8f6c86116acd8566e2da3'
    print(f'condition ids: \n'
          f'{access_id} \n'
          f'{lock_id} \n'
          f'{escrow_id}')
    agreement_hash = ServiceAgreement.generate_service_agreement_hash(
        sa.template_id,
        (access_id, lock_id, escrow_id),
        sa.conditions_timelocks,
        sa.conditions_timeouts,
        agreement_id,
        generate_multi_value_hash
    )
    print('agreement hash: ', agreement_hash.hex())
    expected = '0x96732b390dacec0f19ad304ac176b3407968a0184d01b3262687fd23a3f0995e'
    print('expected hash: ', expected)
    assert agreement_hash.hex() == expected, 'hash does not match.'

# THIS TEST IS ONLY VALID FOR SIMULATING HASH ENCODINGS IN A FIXED ENVIRONMENT
# USEFUL TO VALIDATE DIFFERENCES BETWEEN PARAMETER ENCODING BETWEEN SDK IMPLEMENTATIONS
#
# def test_escrow_payment_condition_id(setup_agreements_fixed_environment):
#     """Test the escrow reward condition id"""
#
#
#     (
#         keeper,
#         ddo,
#         publisher_acc,
#         consumer_acc,
#         agreement_id,
#         asset_id,
#         price,
#         amounts,
#         receivers,
#         service_agreement,
#         (lock_cond_id, access_cond_id, escrow_cond_id),
#
#     ) = setup_agreements_fixed_environment
#
#     checksum_addresses = to_checksum_addresses(receivers)
#
#     # Give consumer some tokens
#     keeper.dispenser.request_vodkas(price, consumer_acc)
#
#     (access_cond_id,
#      lock_cond_id,
#      escrow_cond_id) = service_agreement.generate_agreement_condition_ids(
#         agreement_id, asset_id, consumer_acc.address, publisher_acc.address, keeper
#     )
#     escrow_cond_id_generated = keeper.escrow_payment_condition.generate_id(
#         agreement_id,
#         service_agreement.condition_by_name['escrowPayment'].param_types,
#         [amounts, checksum_addresses, publisher_acc.address, lock_cond_id, access_cond_id]
#     ).hex()
#
#     assert escrow_cond_id == escrow_cond_id_generated
#
#     success = keeper.access_template.create_agreement(
#         agreement_id,
#         asset_id,
#         [access_cond_id, lock_cond_id, escrow_cond_id],
#         service_agreement.conditions_timelocks,
#         service_agreement.conditions_timeouts,
#         consumer_acc.address,
#         consumer_acc
#     )
#     print('create agreement: ', success)
#     assert success, f'createAgreement failed {success}'
#     event = keeper.access_template.subscribe_agreement_created(
#         agreement_id,
#         10,
#         log_event(keeper.access_template.AGREEMENT_CREATED_EVENT),
#         (),
#         wait=True
#     )
#     assert event, 'no event for AgreementCreated '
#
#     # Verify condition types (condition contracts)
#     agreement = keeper.agreement_manager.get_agreement(agreement_id)
#     assert agreement.did == asset_id, ''
#     cond_types = keeper.access_template.get_condition_types()
#     for i, cond_id in enumerate(agreement.condition_ids):
#         cond = keeper.condition_manager.get_condition(cond_id)
#         assert cond.type_ref == cond_types[i]
#         assert int(cond.state) == 1
#
#     # Give consumer some tokens
#     keeper.dispenser.request_vodkas(price, consumer_acc)
#
#     # Fulfill lock_payment_condition
#     pub_token_balance = keeper.token.get_token_balance(publisher_acc.address)
#     starting_balance = keeper.token.get_token_balance(keeper.escrow_payment_condition.address)
#     keeper.token.token_approve(keeper.lock_payment_condition.address, price, consumer_acc)
#     tx_hash = keeper.lock_payment_condition.fulfill(
#         agreement_id, keeper.escrow_payment_condition.address, price, consumer_acc)
#     keeper.lock_payment_condition.get_tx_receipt(tx_hash)
#     event = keeper.lock_payment_condition.subscribe_condition_fulfilled(
#         agreement_id,
#         10,
#         log_event(keeper.lock_payment_condition.FULFILLED_EVENT),
#         (),
#         wait=True
#     )
#     assert event, 'no event for LockPaymentCondition.Fulfilled'
#     assert keeper.condition_manager.get_condition_state(lock_cond_id) == 2, ''
#     assert keeper.token.get_token_balance(
#         keeper.escrow_payment_condition.address
#     ) == (price + starting_balance), ''
#
#     # Fulfill access_condition
#     tx_hash = keeper.access_condition.fulfill(
#         agreement_id, asset_id, consumer_acc.address, publisher_acc
#     )
#     keeper.access_condition.get_tx_receipt(tx_hash)
#     event = keeper.access_condition.subscribe_condition_fulfilled(
#         agreement_id,
#         20,
#         log_event(keeper.access_condition.FULFILLED_EVENT),
#         (),
#         wait=True
#     )
#     assert event, 'no event for AccessCondition.Fulfilled'
#     assert keeper.condition_manager.get_condition_state(access_cond_id) == 2, ''
#
#
#     # Fulfill escrow_payment_condition
#     tx_hash = keeper.escrow_payment_condition.fulfill(
#         agreement_id, amounts, checksum_addresses,
#         publisher_acc.address, lock_cond_id,
#         access_cond_id, publisher_acc
#     )
#     keeper.escrow_payment_condition.get_tx_receipt(tx_hash)
#     assert keeper.condition_manager.get_condition_state(escrow_cond_id) == 2, ''
#
#     print('agreementId: ' + agreement_id)
#     print('did: ' + asset_id)
#     print('lockConditionId: ' + lock_cond_id)
#     print('accessConditionId: ' + access_cond_id)
#     print('escrowConditionId: ' + escrow_cond_id)
#     print('publisherAccount: ' + publisher_acc.address)
#     print('consumerAccount: ' + consumer_acc.address)


def test_agreement():
    template_id = Web3.toChecksumAddress('0x' + ('f' * 40))
    agreement_id = '0x' + ('e' * 64)
    access_id = '0x' + ('a' * 64)
    lock_id = '0x' + ('b' * 64)
    escrow_id = '0x' + ('c' * 64)

    signature = ServiceAgreement.generate_service_agreement_hash(
        template_id,
        [access_id, lock_id, escrow_id],
        [0, 0, 0],
        [0, 0, 0],
        agreement_id,
        generate_multi_value_hash
    )

    print({signature})
    assert (
            signature == Web3.toBytes(
        hexstr="0x67901517c18a3d23e05806fff7f04235cc8ae3b1f82345b8bfb3e4b02b5800c7"
    )), "The signature is not correct."



from eth_utils import add_0x_prefix, remove_0x_prefix
from web3 import Web3


class Parameter:
    """
    Parameter of the condition.
    Form by a parameter name, a type and a value.
    """

    def __init__(self, param_json):
        self.name = param_json['name']
        self.type = param_json['type']
        self.value = param_json['value']
        if self.type == 'bytes32':
            self.value = add_0x_prefix(self.value)

    def as_dictionary(self):
        """
        Return the parameter as a dictionary.

        :return: dict
        """
        return {
            "name": self.name,
            "type": self.type,
            "value": remove_0x_prefix(self.value) if self.type == 'bytes32' else self.value
        }


class Event:
    """
    Example: (formatted to make Sphinx happy!)

    {
    "name": "PaymentLocked",
    "actorType": ["publisher"],
    "handlers": {
    "moduleName": "accessControl",
    "functionName": "grantAccess",
    "version": "0.1"
    }
    }
    """

    def __init__(self, event_json):
        self.values_dict = dict(event_json)

    @property
    def name(self):
        return self.values_dict['name']

    @property
    def actor_type(self):
        return self.values_dict['actorType']

    @property
    def handler_module_name(self):
        return self.values_dict['handler']['moduleName']

    @property
    def handler_function_name(self):
        return self.values_dict['handler']['functionName']

    @property
    def handler_version(self):
        return self.values_dict['handler']['version']

    def as_dictionary(self):
        return self.values_dict


class ServiceAgreementCondition(object):
    """Class representing the Service Agreement Conditions."""

    def __init__(self, condition_json=None):
        self.name = ''
        self.timelock = 0
        self.timeout = 0
        self.contract_name = ''
        self.function_name = ''
        self.is_terminal = False
        self.dependencies = []
        self.timeout_flags = []
        self.parameters = []
        self.events = []
        if condition_json:
            self.init_from_condition_json(condition_json)

    def init_from_condition_json(self, condition_json):
        """
        Init the condition values from a condition json.

        :param condition_json: dict
        """
        self.name = condition_json['name']
        self.timelock = condition_json['timelock']
        self.timeout = condition_json['timeout']
        self.contract_name = condition_json['contractName']
        self.function_name = condition_json['functionName']
        self.parameters = [Parameter(p) for p in condition_json['parameters']]
        self.events = [Event(e) for e in condition_json['events']]

    def as_dictionary(self):
        """
        Return the condition as a dictionary.

        :return: dict
        """
        condition_dict = {
            "name": self.name,
            "timelock": self.timelock,
            "timeout": self.timeout,
            "contractName": self.contract_name,
            "functionName": self.function_name,
            "events": [e.as_dictionary() for e in self.events],
            "parameters": [p.as_dictionary() for p in self.parameters],
        }

        return condition_dict

    @property
    def param_types(self):
        """
        Type of the conditions.

        :return: list of types.
        """
        return [parameter.type for parameter in self.parameters]

    @property
    def param_values(self):
        """
        Values of the conditions.

        :return: list of values.
        """
        return [parameter.value for parameter in self.parameters]

    @property
    def values_hash(self):
        """
        Value hashes

        :return:
        """
        return Web3.soliditySha3(self.param_types, self.param_values).hex()



import requests
from requests.adapters import HTTPAdapter


def get_requests_session():
    """
    Set connection pool maxsize and block value to avoid `connection pool full` warnings.

    :return: requests session
    """
    session = requests.sessions.Session()
    session.mount('http://', HTTPAdapter(pool_connections=25, pool_maxsize=25, pool_block=True))
    session.mount('https://', HTTPAdapter(pool_connections=25, pool_maxsize=25, pool_block=True))
    return session



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""The setup script."""
import os
from os.path import join

from setuptools import setup

with open('README.md') as readme_file:
    readme = readme_file.read()

with open('CHANGELOG.md') as history_file:
    history = history_file.read()

# Installed by pip install common-utils-py
# or pip install -e .
install_requirements = [
    'requests==2.21.0',
    'eciespy==0.3.5',
    'eth-keys==0.3.3',
    'rsa==4.7',
    # secp256k1 support was added recently and the latest release does not included it yet
    # so for now we are going to used a fork
    'nevermined-authlib==0.1.0'
]

# Required to run setup.py:
setup_requirements = ['pytest-runner', ]

test_requirements = [
    'contracts-lib-py==0.11.0',
    'web3==5.26.0',
    'coverage',
    'docker',
    'mccabe',
    'pylint',
    'pytest',
    'pytest-watch',
]

# Possibly required by developers of common-utils-py:
dev_requirements = [
    'bumpversion',
    'pkginfo',
    'twine',
    'watchdog',
]

docs_requirements = [
    'Sphinx',
    'sphinxcontrib-apidoc',
]

packages = []
for d, _, _ in os.walk('common_utils_py'):
    if os.path.exists(join(d, '__init__.py')):
        packages.append(d.replace(os.path.sep, '.'))

setup(
    author="nevermined-io",
    author_email='root@nevermined.io',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3.7',
    ],
    description="🐳 Library including all the common functionalities used in Python projects",
    extras_require={
        'test': test_requirements,
        'dev': dev_requirements + test_requirements + docs_requirements,
        'docs': docs_requirements,
    },
    install_requires=install_requirements,
    license="Apache Software License 2.0",
    long_description=readme,
    long_description_content_type="text/markdown",
    include_package_data=True,
    keywords='common-utils-py',
    name='common-utils-py',
    packages=packages,
    setup_requires=setup_requirements,
    test_suite='tests',
    tests_require=test_requirements,
    url='https://github.com/nevermined-io/common-utils-py',
    version='0.10.0',
    zip_safe=False,
)
